/*
 * @(#) AlgorithmRVM.java  1.10 02/09/03
 *
 * Copyright ***,  All Rights Reserved.
 *
 * This software is the proprietary information of ********
 * Use is subject to license terms.
 *
 */

// import java packages
//
import java.awt.*;
//These imports are not needed - Phil T. 6-23-03
//import java.awt.event.*;
//import javax.swing.*;
//import javax.swing.border.*;
//import javax.swing.event.*;
//import java.io.*;
//import java.lang.*;
import java.util.*;

/**
 * This interface is designed to be the base for all algorithms
 * .....
 * .....
 * @version 1.00
 */
public class AlgorithmRVM
    extends Algorithm {

  //-----------------------------------------------------------------
  //
  // static data members
  //
  //-----------------------------------------------------------------

  static final double DEF_ALPHA_THRESH = 1e12;
  static final double DEF_MIN_ALLOWED_WEIGHT = 1e-12;
  static final long DEF_MAX_RVM_ITS = (long) (1 << 30);
  static final long DEF_MAX_UPDATE_ITS = (long) (1 << 30);
  static final double DEF_MIN_THETA = 1e-8;
  static final double DEF_MOMENTUM = 0.85;
  static final long DEF_MAX_ADDITIONS = 1;
  static final boolean DEF_SAVE_RESTART = false;
  static final boolean DEF_LOAD_RESTART = false;

  static final int KERNEL_TYPE_LINEAR = 0;
  static final int KERNEL_TYPE_RBF = 1;
  static final int KERNEL_TYPE_POLYNOMIAL = 2;
  static final int KERNEL_TYPE_DEFAULT = KERNEL_TYPE_LINEAR;

  static final Double CLASS_LABEL_1 = new Double(0.0);
  static final Double CLASS_LABEL_2 = new Double(1.0);

  //-----------------------------------------------------------------
  //
  // instance data members
  //
  //-----------------------------------------------------------------

  boolean debug_level_d = false;

  // RVM data member
  //
  // relevance vector weights, vectors, and labels
  //
  Vector x_d = new Vector(); // vectors_d
  Vector y_d = new Vector(); // targets_d
  Vector evalx_d = new Vector(); // vectors_d
  Vector evaly_d = new Vector(); // targets_d

  Matrix inv_hessian_d = new Matrix(); // A in [2]
  Vector weights_d = new Vector(); // w in [1]
  double bias_d = 0.0;

  // RVM parameters
  //
  //int kernel_type_d = KERNEL_TYPE_LINEAR;
  int kernel_type_d = KERNEL_TYPE_RBF;

  // vector of support region points
  //
  Vector support_vectors_d = new Vector();
  Vector decision_regions_d = new Vector();
  int output_canvas_d[][];

  // tuning parameters: These are the only parameters that a user need
  // worry about prior to training. The default quantities are usually
  // sufficient. However, run-time performance and accuracy can be
  // influenced by appropriately tuning these parameters.
  //
  // maximum hyperparameter value allowed before pruning. decreasing this
  // value can speed up convergence of the model but may yield overpruning
  // and poor generalization. the value should always be greater than zero
  //
  double alpha_thresh_d;

  // minimum value of a weight allowed in the model. typically as the weight
  // decreases toward zero, it should be pruned.
  //
  double min_allowed_weight_d;

  // maximum number of training iterations to carry out before stopping
  // adjusting this parameter can result in sub-optimal results
  //
  long max_rvm_its_d;

  // maximum number of iterations that are allowed to pass betweeen
  // model updates (adding or pruning of a hyperparameter) before training
  // is terminated (for the full mode of training) or a vector is manually
  // added (for the incremental mode of training)
  //
  long max_update_its_d;

  // minimum value of the theta calculation (the divisor of equation
  // 17 in [3]) that will trigger a model addition (in the
  // incremental training mode).
  //
  double min_theta_d;

  // hyperparameter update momentum term. a larger value for this term
  // can lead to faster convergence, while too large a value can cause
  // oscillation. the value is typically on the range [0,1]
  //
  double momentum_d;

  // number of hyperparameters to add at a time. adding a small number of
  // hyperparameters at a time will yield a smoother movement through the
  // model space, but may increase the total convergence time.
  //
  long max_additions_d;

  // whether or not to create backup copies of training data. if true then
  // data will be occasionally saved to disk in the file provided. that
  // file can later be used to restart training in the middle of the
  // convergence process. *** the restart facility currently is available
  // only for incremental training ***
  //
  boolean save_restart_d;
  //Filename restart_save_file_d;

  // whether or not to bootstrap training from a restart file. if true then
  // the given restart file is read and training is continued from that point
  // forward. *** the restart facility currently is available
  // only for incremental training ***
  //
  boolean load_restart_d;
  //Filename restart_load_file_d;

  // model data
  //
  int num_samples_d; // number of remaining RVs
  Matrix A_d = new Matrix(); // hyperparameter matrix
  int dimA_d; // number of non-pruned params
  Matrix phi_d = new Matrix(); // working design matrix
  Vector curr_weights_d = new Vector(); // updated weights
  Vector last_rvm_weights_d = new Vector(); // stored weights for rvm pass

  // IRLS training quantities
  //
  Vector sigma_d = new Vector(); // error vector
  Matrix B_d = new Matrix(); // data-dependent "noise"
  Vector gradient_d = new Vector(); // gradient w.r.t. weights
  Matrix hessian_d = new Matrix(); // hessian w.r.t. weights
  Matrix covar_cholesky_d = new Matrix(); // cholesky decomposition of covar
  Vector old_irls_weights_d = new Vector(); // stored weights for irls pass
  long last_changed_d; // counter for last time model
  //   changed

  // incremental training quantities
  //
  Vector S_d = new Vector(); // updates for incremental train
  //Vector hyperparams_d = new Vector(); // current hyperparameters
  //Vector weights_d;                  // current hyperparameters
  Vector last_hyperparams_d = new Vector(); // previous iterations hyperparams
  Vector twoback_hyperparams_d = new Vector(); // hyperparameters from two

  // classification functions
  //
  public boolean initialize() {
    // Debug
    // System.out.println("AlgorithmRVM : initialize()");

    // check the data points
    //
    if (output_panel_d == null) {
      return false;
    }

    alpha_thresh_d = 1e4;
    min_allowed_weight_d = 1e-8;
    max_rvm_its_d = DEF_MAX_RVM_ITS;
    max_update_its_d = DEF_MAX_UPDATE_ITS;
    max_update_its_d = 100;
    min_theta_d = DEF_MIN_THETA;
    momentum_d = DEF_MOMENTUM;
    max_additions_d = DEF_MAX_ADDITIONS;
    save_restart_d = DEF_SAVE_RESTART;
    load_restart_d = DEF_LOAD_RESTART;

    // add the process description for the RVM algorithm
    //
    if (description_d.size() == 0) {
      String str = new String("   0. Initialize the original data.");
      description_d.addElement(str);

      str = new String("   1. Displaying the original data.");
      description_d.addElement(str);

      str = new String("   2. Computing the Relevance Vectors.");
      description_d.addElement(str);

      str = new String("   3. Computing the decision regions.");
      description_d.addElement(str);
    }

    // append message to process box
    //
    pro_box_d.appendMessage("Relevance Vector Machine :" + "\n");

    // set the data points for this algorithm
    //
    set1_d = (Vector) data_points_d.dset1.clone();
    set2_d = (Vector) data_points_d.dset2.clone();

    // reset values
    //
    support_vectors_d = new Vector();
    decision_regions_d = new Vector();
    step_count = 3;
    x_d = new Vector();
    y_d = new Vector();

    // set the step index
    //
    step_index_d = 0;

    // append message to process box
    //
    pro_box_d.appendMessage( (String) description_d.get(step_index_d));

    // exit gracefully
    //
    return true;
  }

  // the Runnable method
  //
  public void run() {
    // Debug
    // System.out.println(algo_id + ": run()");

    if (step_index_d == 1) {
      disableControl();
      step1();
      enableControl();
    }

    else if (step_index_d == 2) {
      disableControl();
      step2();
      enableControl();
    }

    else if (step_index_d == 3) {
      disableControl();
      step3();
      pro_box_d.appendMessage("   Step Sequence Complete");
      enableControl();
    }

    // exit gracefully
    //
    return;
  }

  boolean step1() {

    // System.out.println(algo_id + ": step1()");

    pro_box_d.setProgressMin(0);
    pro_box_d.setProgressMax(1);
    pro_box_d.setProgressCurr(0);

    scaleToFitData();

    // Display original data
    output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT,
                             data_points_d.color_dset1);
    output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
                             data_points_d.color_dset2);
    output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
                             data_points_d.color_dset3);
    output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT,
                             data_points_d.color_dset4);

    // step 1 completed
    //
    pro_box_d.setProgressCurr(1);
    output_panel_d.repaint();

    // exit gracefully
    //
    return true;
  }

  // method: step2
  //
  // arguments: none
  // return   : none
  //
  // step one of the algorithm
  //
  boolean step2() {
    // Debug
    // System.out.println("AlgorithmRVM : step2()");

    //sequentialMinimalOptimization();

    pro_box_d.setProgressMin(0);
    pro_box_d.setProgressMax(4) ;
    pro_box_d.setProgressCurr(0);

    trainFull();
    //computeSupportVectors();

    // display support vectors
    //
    //color.addElement(new Color(255, 200, 0));    // orange
    output_panel_d.addOutput(support_vectors_d, Classify.PTYPE_SUPPORT_VECTOR,
                             Color.black);
    //new Color(255, 200, 0));
    pro_box_d.setProgressCurr(20);
    output_panel_d.repaint();

    /*
      // append message to process box
      //
      //transformRVM(data_points_d);
      //computeMeans();
      //displayMatrices();
      //displaySupportRegions();
      sequentialMinimalOptimization();
      computeSupportVectors();
      // display support vectors
      //
      //color.addElement(new Color(255, 200, 0));    // orange
      output_panel_d.addOutput(support_vectors_d,
      Classify.PTYPE_SUPPORT_VECTOR,
      Color.black);
      //new Color(255, 200, 0));
      output_panel_d.repaint();
     */

    // exit gracefully
    //
    return true;
  }

  // method: step3
  //
  // arguments: none
  // return   : none
  //
  // step one of the algorithm
  //
  boolean step3() {
    // Debug
    // System.out.println("AlgorithmRVM : step3()");
    computeDecisionRegions();

    // display support vectors
    //
    output_panel_d.addOutput(decision_regions_d, Classify.PTYPE_INPUT,
                             new Color(255, 200, 0));
    //Color.black);
    output_panel_d.repaint();

    computeErrors();

    /*
      //computeDecisionRegions();
      // display support vectors
      //
      //output_panel_d.addOutput(decision_regions_d, Classify.PTYPE_INPUT,
      //new Color(255, 200, 0));
      //Color.black);
      //output_panel_d.repaint();
     */

    // exit gracefully
    //
    return true;
  }

  // method: trainFull
  //
  // argument:
  //  RVMTrainData&: (input) training data object
  //
  // return: boolean value indicating status
  //
  // this method trains an RVM probabilistic classifier on the input data and
  // targets provided. the inputs and targets should have a one-to-one
  // correspondence and all targets should be either 0 (out-of-class) or
  // 1 (in-class). The training scheme follows that of [1] section 3.
  // It is assumed that the class data and targets are already set when
  // this method is called.
  //
  // the training algorithm in [1] for RVMs proceeds in three iterative steps
  //
  //   1. prune away any weights whose hyperparameters have gone to infinity
  //
  //   2. estimate most probable weights: in this step we find those
  //   weights that maximize equation (24) of [1].  The iteratively
  //   reweighted least squares algorithm is used to find w_mp
  //
  //   3. estimate the covariance of a Gaussian approximation to the
  //   posterior distribution (the posterior is what we want to
  //   model in the end) over the weights centered at the weights,
  //   w_mp.
  //
  //   4. estimate the hyperparameters that govern the weights. this
  //   is done by evaluating the derivative over the hyperparameters
  //   and finding the maximizing hyperparameters.
  //
  // 1, 2, 3, and 4 are carried out iteratively until a suitable convergence
  // criteria is satisfied.
  //
  public boolean trainFull() {
    // Debug
    // System.out.println("AlgorithmRVM : trainFull()");

    // debugging information
    //
    //debug_level_d = true;

    // 0. initialize data structures for training
    //
    if (!initFullTrain()) {
      // System.out.println("Error at initFullTrain ");
    }

    if (debug_level_d) {
      // System.out.println("RVM training");
    }

    if (debug_level_d) {
      //Matrix.printDoubleVector(x_d);
      Matrix.printDoubleVector(y_d);
    }

    // debugging information
    //
    if (debug_level_d) {
      // System.out.println("RVM convergence loop");
    }

    // loop until convergence or until a maximum number of iterations has
    // been reached
    //
    long num_rvm_its = 0;
    boolean rvm_converged = false;
    while (!rvm_converged) {

      // store the weights achieved on the last iteration so we
      // can test for convergence later
      //
      last_rvm_weights_d = (Vector) curr_weights_d.clone();

      // debugging information
      //
      if (debug_level_d) {
        // System.out.println("RVM Iteration: " + num_rvm_its);
      }

      if (debug_level_d) {
        // System.out.println("phi_d: ");
        phi_d.printMatrix();
        // System.out.println("A_d: ");
        A_d.printMatrix();
        // System.out.println("weights_d: ");
        Matrix.printDoubleVector(weights_d);
        // System.out.println("curr_weights_d: ");
        Matrix.printDoubleVector(curr_weights_d);
      }

      // 1. prune only after the first iteration (we use a
      // weight of exactly 0.0 to trigger pruning so pruning on
      // the first iteration would prune all weights
      //
      if (num_rvm_its > 0) {
        pruneAndUpdate();
      }

      if (debug_level_d) {
        // System.out.println("After pruning: ");
        // System.out.println("phi_d: ");
        phi_d.printMatrix();
        // System.out.println("A_d: ");
        A_d.printMatrix();
        // System.out.println("weights_d: ");
        Matrix.printDoubleVector(weights_d);
        // System.out.println("curr_weights_d: ");
        Matrix.printDoubleVector(curr_weights_d);
        //// System.out.println("curr_weights_d: ");
        //Matrix.printMatrix(curr);
      }

      // if all weights have been pruned, then there is nothing
      // left to do and the process has converged (albeit to a
      // pretty useless solution)
      //
      if (dimA_d == 0) {
        rvm_converged = true;

        // debugging information
        //
        if (debug_level_d) {
          // System.out.println("rvm convergence achieved");
        }

        // conclude training
        //
        break;
      }

      // 2. run a pass of IRLS training to estimate w_MP
      //
      irlsTrain();

      // 3. estimate the covariance of the Gaussian
      // approximation compute the variance vector. The
      // covariance is the inverse of the Hessian matrix. Only
      // the diagonal elements are needed to update the
      // hyperparameters. From the Cholesky decomposition, we
      // can efficiently find these values. After this function
      // call, the diagonal elements of covar_cholesky will
      // contain the negation of the diagonal elements of the
      // estimated covariance.  The other elements of
      // covar_cholesky are not meaningful.
      //
      computeVarianceCholesky();

      // 4. reestimate the hyperparameters
      //
      boolean A_changed = false;
      A_changed = updateHyperparametersFull();

      // store the final weights from this iteration
      //
      int start_index = 0;
      if (num_samples_d < dimA_d) {
        start_index = 1;
        bias_d = MathUtil.doubleValue(curr_weights_d, 0);
      }
      MathUtil.copyVector(weights_d, curr_weights_d, num_samples_d, 0,
                          start_index);

      /*
          // debugging information
          //
          if (debug_level_d > Integral::DETAILED) {
          curr_weights_d.debug(L"weights after rvm iteration");
          A_d.debug(L"A after rvm iteration");
          }
       */

      // check convergence: converge if the weight updates have
      // stagnated and we have not pruned off a weight
      //
      if (MathUtil.almostEqual(curr_weights_d, last_rvm_weights_d)) {

        last_changed_d++;
        if (!A_changed || (num_rvm_its > max_rvm_its_d) ||
            (last_changed_d > max_update_its_d || weights_d.size() < 2)) {
          rvm_converged = true;
        }
      }
      else {
        last_changed_d = 0;
      }

      /*
          // debugging information
          //
          if (rvm_converged) {
          if (debug_level_d > Integral::NONE) {
        Console::put(L"rvm convergence achieved");
          }
          }
          if (debug_level_d > Integral::NONE) {
          Console::decreaseIndention();
          }
       */

      // next iteration of RVM optimization
      //
      num_rvm_its++;
      //      System.out.println(num_rvm_its);
      //      pro_box_d.setProgressCurr((int)num_rvm_its);

    }

    // run any final steps in the training algorithm and store the
    // final model
    //
    finalizeTraining();

    // debugging information
    //
    if (debug_level_d) {
      // System.out.println(" convergence achieved");
    }

    /*
       // debugging information
       //
       if (debug_level_d > Integral::NONE) {
        Console::decreaseIndention();
        Console::put(L"End RVM training");
        y_d.debug(L"final targets");
        x_d.debug(L"final relevance vectors");
        weights_d.debug(L"final weights");
        bias_d.debug(L"final bias");
       }
     */

    // exit gracefully
    //
    return true;
  }

  // method: initFullTrain
  //
  // argument:
  //  RVMTrainData& tdata: data stored for training the models
  //
  // return: boolean value indicating status
  //
  // initializes the data structures in preparation for a full training
  // pass
  //
  public boolean initFullTrain() {
    // Debug
    // System.out.println("AlgorithmRVM : initFullTrain()");

    // initialize x_d and y_d
    //
    // total number of points
    //
    x_d.clear();
    y_d.clear();
    support_vectors_d.clear();

    /*
       Point p;
       Vector vec_point = new Vector();
       vec_point.add(new Double(-0.5545023696682464));
       vec_point.add(new Double(0.4976303317535544));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_1);
       p = new Point(47,53);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(-0.5450236966824644));
       vec_point.add(new Double(0.2890995260663507));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_1);
       p = new Point(48,75);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(-0.22274881516587675));
       vec_point.add(new Double(0.27014218009478663));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_1);
       p = new Point(82,77);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(-0.21327014218009477));
       vec_point.add(new Double(0.4691943127962085));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_1);
       p = new Point(83,56);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(0.3744075829383888));
       vec_point.add(new Double(-0.31753554502369674));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_2);
       p = new Point(145,139);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(0.6303317535545025));
       vec_point.add(new Double(-0.31753554502369674));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_2);
       p = new Point(172,139);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(0.2606635071090049));
       vec_point.add(new Double(-0.4218009478672986));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_2);
       p = new Point(133,150);
       support_vectors_d.add(p);
       vec_point = new Vector();
       vec_point.add(new Double(0.49763033175355464));
       vec_point.add(new Double(-0.4407582938388628));
       x_d.add(vec_point);
       y_d.add(CLASS_LABEL_2);
       p = new Point(158, 152);
       support_vectors_d.add(p);
       if ( debug_level_d ) {
        Matrix.printDoubleVector(y_d);
       }
     */

    // add the data points
    //
    for (int i = 0; i < set1_d.size(); i++) {
      Vector vec_point = new Vector();
      MyPoint curr_point = (MyPoint) set1_d.get(i);
      support_vectors_d.add(curr_point);

      vec_point.add(new Double(curr_point.x));
      vec_point.add(new Double(curr_point.y));
      x_d.add(vec_point);
      y_d.add(CLASS_LABEL_1);
    }

    //// System.out.println("class: 2");

    for (int i = 0; i < set2_d.size(); i++) {
      Vector vec_point = new Vector();
      MyPoint curr_point = (MyPoint) set2_d.get(i);
      support_vectors_d.add(curr_point);

      vec_point.add(new Double(curr_point.x));
      vec_point.add(new Double(curr_point.y));
      x_d.add(vec_point);
      y_d.add(CLASS_LABEL_2);
    }

    evalx_d = (Vector) x_d.clone();
    evaly_d = (Vector) y_d.clone();

    // number of samples stored
    //
    num_samples_d = x_d.size();
    //// System.out.println(num_samples_d);

    // the weights are a column vector of length N. The bias is stored
    // separately
    //
    weights_d.setSize(num_samples_d);
    MathUtil.initDoubleVector(weights_d, 0.0);

    // the hyperparameters, alphas, are stored as a diagonal array of
    // dimension N+1 x N+1.
    //
    A_d.initMatrixValue(num_samples_d + 1, num_samples_d + 1, 0,
                        Matrix.DIAGONAL);

    // the phi matrix (pp. 214 of [1]) is the design matrix of kernel
    // evaluations. Note that here we use the transpose of the PHI matrix
    // in [1] since our matrices are assumed to be column matrices. Thus
    // phi is an N+1 x N matrix column matrix rather than a N x N+1 row
    // matrix as they do in [1].
    //
    phi_d.initMatrixValue(num_samples_d + 1, num_samples_d, 0, Matrix.FULL);

    // the sigma vector is the probability of each sample given
    // the current weights. the B matrix is a diagonal matrix of
    // size N x N that is used in the least squares optimization
    // of the weights (pp. 219 of [1]).  B[i,i] = sigma(i) * (1 -
    // sigma(i))
    //
    sigma_d.setSize(num_samples_d);
    MathUtil.initDoubleVector(sigma_d, 0.0);
    B_d.initMatrixValue(num_samples_d, num_samples_d, 0, Matrix.DIAGONAL);

    // temporary storage for the gradient vector, hessian matrix and the
    // covariance matrix (or at least the cholesky decomposition of the
    // hessian which can be used to find the covariance)
    //
    gradient_d.setSize(num_samples_d + 1);
    MathUtil.initDoubleVector(gradient_d, 0.0);
    hessian_d.initMatrixValue(num_samples_d + 1, num_samples_d + 1, 0,
                              Matrix.SYMMETRIC);

    covar_cholesky_d.initMatrixValue(num_samples_d + 1, num_samples_d + 1, 0,
                                     Matrix.LOWER_TRIANGULAR);

    // curr_weights are used during the optimization to hold the
    // current set of weights in a single vector. last_rvm_weights
    // hold the last set of weights that were converged to in the
    // RVM training iteration. If these weights match the weights
    // found in the next iteration then we have reached global
    // convergence (subject to a few other rules).
    // old_irls_weights contains the last set of weights tried for
    // the IRLS convergence tests
    //
    curr_weights_d.setSize(num_samples_d + 1);
    MathUtil.initDoubleVector(curr_weights_d, 0.0);
    last_rvm_weights_d.setSize(num_samples_d + 1);
    MathUtil.initDoubleVector(last_rvm_weights_d, 0.0);

    // if the weights stagnate then we can discontinue training even if the
    // hyperparameters are oscillating.
    //
    last_changed_d = 0;

    // compute the phi matrix:
    // The first element in each column is set to 1.0. Element i,j in the
    // matrix corresponds to the kernel function evaluated with
    // x_d(j) and x_d(i-1) as arguments
    //
    double kernel_value = 0.0;
    for (int j = 0; j < num_samples_d; j++) {
      phi_d.setValue(0, j, 1.0);
    }
    for (int i = 1; i <= num_samples_d; i++) {
      for (int j = 0; j < num_samples_d; j++) {
        Vector vec1 = (Vector) x_d.get(j);
        Vector vec2 = (Vector) x_d.get(i - 1);
        kernel_value = K(vec1, vec2);
        phi_d.setValue(i, j, kernel_value);
        //// System.out.println("k_val: " +  kernel_value);
      }
    }

    // initialize the A matrix:
    //  A is initialized to the N+1 x N+1 indentity matrix
    //
    double scale = 1.0 / num_samples_d / num_samples_d;
    A_d.identityMatrix();
    A_d.scalarMultMatrix(scale);

    // keep track of the number of hyperparameters:
    //  if dim_A is equal to the size of the weight_d vector then the bias
    //  has been pruned.
    //
    dimA_d = A_d.getNumRows();

    // initialize weights to zero - this gives every input vector
    // probability 1/2 of being an in-class example.
    // i.e. sigma(0.0) = 1 / (1 + exp(0)) = 1/2
    //
    MathUtil.initDoubleVector(weights_d, 0.0);
    bias_d = 0.0;

    // check that the data and targets are appropriate for training
    //
    if ( (num_samples_d == 0) || (y_d.size() != num_samples_d)) {
      // System.out.println("Error at num_samples_d");
    }

    // exit gracefully
    //
    return true;
  }

  // method: updateHyperparametersFull
  //
  // argument:
  //  boolean& updated: takes a 'true' value if the hyperparameter matrix was
  //                    modified on this iteration - indicating that either
  //                    a hyperparameter went to zero or
  //  RVMTrainData& tdata: data stored for training the models
  //
  // return: boolean value indicating status
  //
  // update the hyperparameter values
  //
  public boolean updateHyperparametersFull() {
    // Debug
    // System.out.println("AlgorithmRVM : updateHyperparametersFull()");

    // update the momentum components
    // only use momentum if no hyperparams have been pruned
    // recently.
    //
    twoback_hyperparams_d = (Vector) last_hyperparams_d.clone();
    A_d.getDiagonalVector(last_hyperparams_d);
    boolean use_momentum = (twoback_hyperparams_d.size() ==
                            last_hyperparams_d.size());

    // loop over each hyperparameter and update - remember that A is
    // diagonal where A(i,i) = alpha_i. The update equation used is
    // equation (16) of [1]. We also test for convergence in this loop.
    // If no alphas have substantially changed then we assume convergence.
    // We must take some care here that NaN and Infinity values do not
    // creap into our calculations
    //
    boolean updated = false;
    double gamma = 0.0;
    double tmp_alpha = 0.0;
    double old_alpha = 0.0;
    double sum_gamma = 0.0;

    for (int i = 0; i < dimA_d; i++) {

      // equation (16): alpha = (1 - alpha(i) * covar(i,i)) / (weights^2)
      //
      old_alpha = A_d.getValue(i, i);

      // error checking
      //
      if (old_alpha < 0.0) {
        // System.out.println("train:invalid old_alpha: " + old_alpha);
        Exception e = new Exception();
        e.printStackTrace();
        return false;
      }

      if (covar_cholesky_d.getValue(i, i) < 0.0) {
        // System.out.println("train:invalid covar");
        Exception e = new Exception();
        e.printStackTrace();
        return false;
      }

      // compute gamma = 1 - alpha(i)*covar(i,i)
      //
      gamma = 1.0 - (old_alpha * covar_cholesky_d.getValue(i, i));

      // gamma can be negative if covar(i,i) ~= 1/old_alpha
      // due to numerical imprecision. only error if the result
      // is not close to zero.
      //
      if (gamma < 0.0) {
        if (!MathUtil.almostEqual(gamma, 0.0)) {
          // System.out.println("train:invalid gamma: " + gamma);
          Exception e = new Exception();
          e.printStackTrace();
          return false;
        }
        else {
          gamma = -gamma;
        }
      }
      sum_gamma += gamma;

      // error checking
      // if the weight is already zero then alpha(i) will go to infinity
      // set it to a very large value instead
      //
      double weight_i = MathUtil.doubleValue(curr_weights_d, i);
      if (weight_i == 0.0) {
        tmp_alpha = alpha_thresh_d * 10;
      }
      else {
        tmp_alpha = gamma / weight_i / weight_i;
      }

      // compute momentum term
      //
      if (use_momentum) {
        double momentum_term = 0.0;
        double last_hyper = MathUtil.doubleValue(last_hyperparams_d, i);
        double last_hyper2 = MathUtil.doubleValue(twoback_hyperparams_d, i);

        if ( (last_hyper < alpha_thresh_d) && (last_hyper2 < alpha_thresh_d)) {

          momentum_term = momentum_d * (last_hyper - last_hyper2);

          if (debug_level_d) {
            // System.out.println("momentum_term = " + momentum_term);
          }

          // the hyperparameter can not be negative so we do not want
          // to decrease toward zero too quickly and we should never
          // allow it to go below zero so we only take moderate
          // negative steps
          //
          if (momentum_term > ( -0.5 * tmp_alpha)) {
            tmp_alpha = tmp_alpha + momentum_term;
          }
        }
      }

      A_d.setValue(i, i, tmp_alpha);

      /*
          // debugging information
          //
          if (debug_level_d > Integral::DETAILED) {
          gamma.debug(L"gamma");
          }
       */

      // zero the weights to be pruned
      //
      if (debug_level_d) {
        // System.out.println("tmp_alpha: " + tmp_alpha + " alpha_thresh_d: " +
	  //                           alpha_thresh_d);
      }

      if (tmp_alpha > alpha_thresh_d) {
        curr_weights_d.set(i, new Double(0.0));
        updated = true;
        if (debug_level_d) {
          // System.out.println("weight set to 0.0");
          Matrix.printDoubleVector(curr_weights_d);
        }
      }

      // check for convergence in A
      //
      if (!MathUtil.almostEqual(tmp_alpha, old_alpha)) {
        updated = true;
      }
    }

    // exit gracefully
    //
    return updated;
  }

  // method: irlsTrain
  //
  // argument:
  //  RVMTrainData& tdata: data stored for training the models
  //
  // return: boolean value indicating status
  //
  // this method completes one pass of IRLS training to update the weights
  // given the currently assigne hyperparamters. In this step, we find
  // those weights that maximize equation (24) of [1].  The iteratively
  // reweighted least squares algorithm is used to find w_mp. It proceeds
  // as follows:
  //
  //   a. initialize
  //   b. loop until convergence
  //      1. compute the B matrix
  //      2. compute the Hessian and the gradient
  //      3. update weights with formula
  //          new weights = weights - inv(Hessian) * gradient
  //      4. check convergence
  //
  public boolean irlsTrain() {
    // Debug
    // System.out.println("AlgorithmRVM : irlsTrain()");

    // 1.a. initialize: curr_weights: holds the weights currently
    // being considered for update
    //
    int start_index = 0;
    if (weights_d.size() < dimA_d) {
      curr_weights_d.set(0, new Double(bias_d));
      start_index = 1;
    }
    MathUtil.copyVector(curr_weights_d, weights_d, weights_d.size(),
                        start_index, 0);

    // make sure there are some weights that need updating
    //
    if (dimA_d == 0) {
      /*
          curr_weights_d.setLength(0);
          old_irls_weights_d.setLength(0);
          gradient_d.setLength(0);
          hessian_d.setDimensions(0,0);
          covar_cholesky_d.setDimensions(0,0);
          sigma_d.assign(0.5);
          B_d.assign(0.25);
       */
      // System.out.println("dimA_d == 0");
      Exception e = new Exception();
      e.printStackTrace();
      return true;
    }

    // 1.b. loop until convergence:
    //   convergence is achieved once the weights do not change over an
    //   iteration. This corresponds to the gradient of the likelihood
    //   function with respect to the weights going to zero.
    //
    boolean irls_converged = false;
    boolean last_iteration = false;
    long num_irls_its = 0;
    double lambda = 1.0;
    boolean reduced = true; // initially true to skip the first iteration
    double old_likelihood = 0.0;
    double new_likelihood = old_likelihood;
    long size_B = B_d.getNumRows();
    Matrix sigma_M = new Matrix();
    Matrix y_M = new Matrix();
    Matrix temp_M = new Matrix();
    Matrix temp2_M = new Matrix();
    Matrix curr_weights_M = new Matrix();
    Matrix gradient_M = new Matrix();

    while (!irls_converged) {

      if (debug_level_d) {
        // System.out.println("curr_weights_d: ");
        Matrix.printDoubleVector(curr_weights_d);
      }

      // copy out the current best weights to the weight vector
      // so they can be used for evaluations
      //
      if (num_irls_its > 0) {
        if (start_index == 1) {
          bias_d = MathUtil.doubleValue(curr_weights_d, 0);
        }
        else {
          bias_d = 0.0;
        }
        MathUtil.copyVector(weights_d, curr_weights_d, weights_d.size(), 0,
                            start_index);
      }

      // 1.b.1. compute the B matrix and the sigma vector
      //
      if (debug_level_d) {
        // System.out.println("IRLS iteration: " + num_irls_its);
        // System.out.println("1.b.1. compute the B matrix and the sigma vector");
      }

      computeSigma();
      for (int i = 0; i < size_B; i++) {
        double sigma = MathUtil.doubleValue(sigma_d, i);
        B_d.setValue(i, i, sigma * (1 - sigma));
      }

      // if the likelihood decreases then the newton step was
      // too large and needs to be reduced
      //
      new_likelihood = computeLikelihood();
      if (debug_level_d) {
        // System.out.println("sigma_d: ");
        Matrix.printDoubleVector(sigma_d);
        // System.out.println("phi_d: ");
        phi_d.printMatrix();
        // System.out.println("B_d: ");
        B_d.printMatrix();
        // System.out.println("old likelihood 1 : " + old_likelihood);
        // System.out.println("new likelihood 1 : " + new_likelihood);
      }

      if (!reduced) {
        if (new_likelihood < old_likelihood || (num_irls_its < 2)) {

          // store the weights computed on the last IRLS iteration
          //
          old_likelihood = new_likelihood;
          if (lambda < 0.9) {
            // System.out.println("increasing step parameter");
            lambda *= 2.0;
          }
          else {
            lambda = 1.0;
          }
        }
        else {

          // revert to the previous weights
          //
          curr_weights_d = (Vector) old_irls_weights_d.clone();
          lambda *= 0.5;
          if (lambda < 0.0001) {
            last_iteration = true;
          }
          reduced = true;
          continue;
        }
      }
      reduced = false;
      old_irls_weights_d = (Vector) curr_weights_d.clone();

      if (debug_level_d) {
        // System.out.println("curr_weights_d: ");
        Matrix.printDoubleVector(curr_weights_d);
        // System.out.println("1.b.2. compute the gradient and Hessian ");
      }

      /* debug
          if (debug_level_d > Integral::BRIEF) {
          Vector error_signal;
          error_signal.sub(targets_d, sigma_d);
          error_signal.abs();
          Double out = error_signal.sum();
          out.debug(L"sum-squared error");
          curr_weights_d.debug(L"current weights");
          }
       */

      // 1.b.2. compute the gradient and Hessian
      //

      // gradient = phi * [y_d - sigma] - A * weights
      // sigma is reused as temporary data space in the second segment
      // for computing A * weights.
      //

      /*
       */

      // init matrix
      //
      boolean status = true;
      sigma_M.initMatrix(sigma_d);
      y_M.initMatrix(y_d);
      curr_weights_M.initMatrix(curr_weights_d);
      gradient_M.initMatrix(gradient_d);

      //
      sigma_M.scalarMultMatrix( -1.0); // -sigma
      sigma_M.addMatrix(y_M); // sigma_M = y_d -sigma_M
      sigma_M.transposeMatrix(temp_M);

      // gradient = phi * [y_d - sigma]
      //
      phi_d.multMatrix(temp_M, gradient_M);

      // temp_M = -A * weights
      curr_weights_M.transposeMatrix(temp_M);
      A_d.multMatrix(temp_M, temp2_M);
      temp2_M.scalarMultMatrix( -1);

      // gradient = phi * [y_d - sigma] - A * weights
      //
      gradient_M.addMatrix(temp2_M);
      //gradient_M.printMatrix();
      gradient_M.toDoubleVector(gradient_d);

      /*
          if (!status) {
          gradient_d.debug(L"computed gradient");
          return Error::handle(name(), L"train:gradient computation", ERR,
            __FILE__, __LINE__);
          }
          // debugging information
          //
          if (debug_level_d > Integral::DETAILED) {
          gradient_d.debug(L"gradient");
          }
       */

      // Hessian = -(phi * B * transpose(phi) + A)
      // this is a quadratic form about B
      //
      status = true;

      // temp_M = phi * B, temp2_M = phi'
      // hessian_d = phi * B * phi'
      //
      phi_d.multMatrix(B_d, temp_M);
      phi_d.transposeMatrix(temp2_M);
      temp_M.multMatrix(temp2_M, hessian_d);

      // phi * B * phi' + A
      //
      hessian_d.addMatrix(A_d);

      /*
          if (!status) {
          hessian_d.debug(L"computed hessian");
          return Error::handle(name(), L"train:Hessian computation", ERR,
            __FILE__, __LINE__);
          }
          // debugging information
          //
          if (debug_level_d > Integral::DETAILED) {
          hessian_d.debug(L"Hessian");
          }
       */

      // with the cholesky decomposition, we can solve systems
      // of equations as well as determine the diagonal elements
      // of the covariance. This is all we need for the RVM
      // training
      //
      if (!hessian_d.decompositionCholesky(covar_cholesky_d)) {
        // System.out.println("train:Cholesky decomposition");
        return false;
      }

      /*
          // debugging information
          //
          if (debug_level_d > Integral::DETAILED) {
          covar_cholesky_d.debug(L"cholesky decomposition");
          }
       */

      if (debug_level_d) {
        // System.out.println("hessian_d: ");
        hessian_d.printMatrix();
        // System.out.println("gradient_d: ");
        Matrix.printDoubleVector(gradient_d);
        // System.out.println("covar_cholesky_d: ");
        covar_cholesky_d.printMatrix();

      }

      // 1.b.3. update the weights:
      //   because the space is convex we take a full Newton step
      //   weights* = weights - inverse(Hessian) * gradient
      //

      // System.out.println("1.b.3. update the weights:");
      status = true;
      gradient_M.initMatrix(gradient_d);
      status &=
          covar_cholesky_d.choleskySolve(temp_M, covar_cholesky_d, gradient_M);

      if (debug_level_d) {
        // System.out.println("temp_M: ");
        temp_M.printMatrix();

      }

      if (!last_iteration) {

        temp_M.scalarMultMatrix(lambda);
        temp2_M.initMatrix(old_irls_weights_d);
        temp_M.addMatrix(temp2_M);
        //temp_M.printMatrix();
        temp_M.toDoubleVector(curr_weights_d);

        double val = MathUtil.vectorProduct(gradient_d, gradient_d);
        val = Math.sqrt(val);
        val /= curr_weights_d.size();

        // 1.b.4. check convergence
        //
        /*
             if ( MathUtil.almostEqual(curr_weights_d, old_irls_weights_d)
             && MathUtil.almostEqual(gradient_d, 0.0 )){
         */
        if (val < 1e-6) {
          last_iteration = true;
        }
      }
      else {
        irls_converged = true;
      }

      /*
          if (!status) {
          covar_cholesky_d.debug(L"computed cholesky");
          return Error::handle(name(), L"train:weight update", ERR,
            __FILE__, __LINE__);
          }
       */

      // next iteration
      //
      num_irls_its++;
      pro_box_d.setProgressCurr( (int)num_irls_its );
      
    }

    if (debug_level_d) {
      // System.out.println("new likelihood 2: " + new_likelihood);
    }

    /*
       if (debug_level_d > Integral::NONE) {
        new_likelihood.debug(L"new likelihood");
       }
     */

    // generate the true hessian (negative of the one we store)
    //
    hessian_d.scalarMultMatrix( -1.0);

    computeSigma();

    /*
       if (debug_level_d > Integral::DETAILED) {
        curr_weights_d.debug(L"current weights after IRLS");
        Vector tmp_1;
        tmp_1.sub(targets_d, sigma_d);
        Vector tmp_weights;
        Matrix tmp_matrix(A_d);
        tmp_matrix.inverse();
        tmp_matrix.changeType(Integral::FULL);
        tmp_matrix.mult(phi_d);
        tmp_matrix.multv(tmp_weights, tmp_1);
        tmp_weights.debug(L"computed weights after IRLS");
       }
     */

    // exit gracefully
    //
    return true;
  }

  // method: finalizeTraining
  //
  // argument:
  //  RVMTrainData& tdata: data stored for training the models
  //
  // return: boolean value indicating status
  //
  // finalize the trained model making it ready to write to file or use
  // in prediction
  //
  boolean finalizeTraining() {
    // Debug
    // System.out.println("AlgorithmRVM : finalizeTraining()");

    // perform a final check on the range of the weights and
    // manually prune any that have fallen below the minimum
    // allowed weight value
    //
    if (Math.abs(bias_d) < min_allowed_weight_d) {
      bias_d = 0.0;
    }
    for (int i = 0; i < weights_d.size(); i++) {
      if (Math.abs(MathUtil.doubleValue(weights_d, i)) < min_allowed_weight_d) {
        weights_d.set(i, new Double(0.0));
      }
    }

    // prune any remaining zero-weights with large hyperparameters
    //
    pruneAndUpdate();

    // run a final iteration of IRLS training to get the final hessian
    //
    irlsTrain();

    // compute the final inverse hessian and assign
    //
    inv_hessian_d.invertMatrix(hessian_d);
    inv_hessian_d.scalarMultMatrix( -1);

    // exit gracefully
    //
    return true;

  }

  // method: pruneAndUpdate
  //
  // argument:
  //  RVMTrainData& tdata: data stored for training the models
  //
  // return: boolean value indicating status
  //
  // prunes off vectors whose hyperparameters have gone to infinity and
  // updates working data sets
  //
  public boolean pruneAndUpdate() {
    // Debug
    // System.out.println("AlgorithmRVM : pruneAndUpdate()");

    /*
       // debugging information
       //
       if (debug_level_d > Integral::BRIEF) {
        A_d.debug(L"A");
       }
     */

    if (debug_level_d) {
      // System.out.println("A_d: ");
      A_d.printMatrix();
      // System.out.println("weights_d: ");
      Matrix.printDoubleVector(weights_d);
      // System.out.println("alpha_thresh_d: " + alpha_thresh_d);
    }

    // prune any weights that have gone to zero
    //
    pruneWeights();

    // update the number of hyperparameters after pruning
    //
    dimA_d = A_d.getNumRows();

    // reset the size of the vector set
    //
    num_samples_d = weights_d.size();

    gradient_d.setSize(dimA_d);
    MathUtil.initDoubleVector(gradient_d, 0.0);
    curr_weights_d.setSize(dimA_d);
    MathUtil.initDoubleVector(curr_weights_d, 0.0);
    hessian_d.initMatrixValue(dimA_d, dimA_d, 0, Matrix.SYMMETRIC);
    covar_cholesky_d.initMatrixValue(dimA_d, dimA_d, 0, Matrix.LOWER_TRIANGULAR);

    // exit gracefully
    //
    return true;
  }

  // method: pruneWeights
  //
  // arguments:
  //  RVMTrainData& tdata: (input) training data storage
  //
  // return: boolean value indicating status
  //
  // this method prunes off vectors which attain a zero weight during
  // training. Auxiliary data structures that are indexed according to the
  // working set are also updated.
  //
  public boolean pruneWeights() {
    // Debug
    // System.out.println("AlgorithmRVM : pruneWeights()");

    // declare local variables
    //
    int num_weights = curr_weights_d.size();

    // the A matrix is diagonal so we can get a temporary version of it
    // into a vector - this makes it easier to work with
    //
    Vector tmp_A = new Vector();
    A_d.getDiagonalVector(tmp_A);

    if (debug_level_d) {
      // System.out.println("tmp_A before pruning:");
      Matrix.printDoubleVector(tmp_A);
    }

    /*
       //A_d.getDiagonal(tmp_A);
       for (int i = 0; i < dimA_d; i++) {
        // prune any weight which has been floored. we prune off
        // the corresponding hyperparameter as well as the row of
        // the phi matrix corresponding to this weight
        //
        if (!((MathUtil.doubleValue(curr_weights_d, i) == 0.0) &&
     (A_d.getValue(i,i) > alpha_thresh_d))){
     flags[i] = true;
     tmp_A.add(new Double(A_d.getValue(i, i)));
     dimA ++;
        }
        else {
     flags[i] = false;
        }
       }
     */

    // get the dimensions of the phi matrix
    //
    int num_rows = phi_d.getNumRows();

    boolean flags[] = new boolean[num_rows];
    int dimA = 0;

    // loop over the weights and check for zero-valued weights.
    // the weights will be exactly zero because we have manually
    // floored them previously
    //
    int offset = weights_d.size() - num_weights;

    int weights_pruned = 0;
    for (int i = 0; i < num_weights; i++) {

      double weights_i = MathUtil.doubleValue(curr_weights_d, i);
      int a_i = i + weights_pruned;
      // prune any weight that is floored
      //
      if ( (weights_i == 0.0) && (A_d.getValue(a_i, a_i) > alpha_thresh_d)) {

        /*
             if (debug_level_d > Integral::BRIEF) {
            Console::put(L"pruning");
             }
         */

        weights_pruned++;

        // remove the weight, target and vector
        //
        if (i + offset >= 0) {
          weights_d.remove(i + offset);
          support_vectors_d.remove(i + offset);
          evalx_d.remove(i + offset);
          evaly_d.remove(i + offset);
        }
        curr_weights_d.remove(i);
        tmp_A.remove(i);

        // because we have removed an element we need to stay at this
        // entry index and update the stop condition
        //
        i = i - 1;
        num_weights = num_weights - 1;
        flags[i + weights_pruned] = false;
      }
      else {
        flags[i + weights_pruned] = true;
        dimA++;
      }
    }

    /*
       // debugging information
       //
       if (debug_level_d > Integral::DETAILED) {
        Console::put(L"Pruning completed");
        bias_d.debug(L"bias_d");
        weights_d.debug(L"weights_d");
        targets_d.debug(L"targets_d");
        vectors_d.debug(L"vectors_d");
       }
     */

    if (debug_level_d) {
      // System.out.println("tmp_A after pruning:");
      Matrix.printDoubleVector(tmp_A);
    }

    dimA_d = dimA;
    A_d.initDiagonalMatrix(tmp_A);
    Matrix tmp_phi = new Matrix();
    tmp_phi.copyMatrixRows(phi_d, flags);
    phi_d = tmp_phi;

    // debugging information
    //
    if (debug_level_d) {
      // System.out.println("Pruning " + weights_pruned + " vectors..." +
	//                         weights_d.size() + " remain");
    }

    // exit gracefully
    //
    return true;
  }

  // method: computeVarianceCholesky
  //
  // arguments:
  //  RVMTrainData& tdata: (input) training data storage
  //
  // return: logical value indicating status
  //
  // this method computes the diagonal elements of the inverse from the
  // cholesky decomposition
  //
  public boolean computeVarianceCholesky() {
    // Debug
    // System.out.println("AlgorithmRVM : computeVarianceCholesky()");
    // find the size of the cholesky decomposition matrix
    //
    long nrows = covar_cholesky_d.getNumRows();
    double sum = 0.0;

    // compute the inverse of the lower triangular cholesky decomposition
    //
    for (int i = 0; i < nrows; i++) {
      covar_cholesky_d.setValue(i, i, (1.0 / covar_cholesky_d.getValue(i, i)));
      for (int j = i + 1; j < nrows; j++) {
        sum = 0.0;
        for (int k = i; k < j; k++) {
          sum -= covar_cholesky_d.getValue(j, k) *
              covar_cholesky_d.getValue(k, i);
        }
        covar_cholesky_d.setValue(j, i, (sum / covar_cholesky_d.getValue(j, j)));
      }
    }

    // loop over columns and get the sumSquare values to put on
    // the diagonals.  This simulates a matrix multiply but we
    // only retain the diagonal values.
    //
    /*
       Vector tmp;
       for (long i = 0; i < nrows; i++) {
        covar_cholesky_d.getColumn(tmp, i);
        double val = tmp.sumSquare();
        covar_cholesky_d.setValue(i,i,val);
       }
     */
    for (int i = 0; i < nrows; i++) {
      double val = covar_cholesky_d.getColSumSquare(i);
      covar_cholesky_d.setValue(i, i, val);
    }

    // exit gracefully
    //
    return true;
  }

  // method: computeSigma
  //
  // arguments:
  //  const RVMTrainData& tdata: (input) training data storage
  //
  // return: boolean value indicating status
  //
  // this method computes the sigma valued defined on pp. 218 of [1].
  //
  public boolean computeSigma() {
    // Debug
    // System.out.println("AlgorithmRVM : computeSigma()");
    // declare local variables
    //
    Matrix weights = new Matrix();
    Matrix sigma = new Matrix();
    //debug_level_d = false;

    weights.initMatrix(curr_weights_d);

    if (debug_level_d) {
      // System.out.println("phi_d: ");
      phi_d.printMatrix();
      // System.out.println("curr_weights_d: ");
      weights.printMatrix();
      // System.out.println("sigma: ");
      sigma.printMatrix();
    }

    weights.multMatrix(phi_d, sigma);

    sigma.scalarMultMatrix( -1);
    sigma.expMatrix();
    sigma.addMatrixElements(1.0);
    sigma.inverseMatrixElements();

    if (debug_level_d) {
      // System.out.println("phi_d: ");
      phi_d.printMatrix();
      // System.out.println("curr_weights_d: ");
      weights.printMatrix();
      // System.out.println("sigma: ");
      sigma.printMatrix();
    }

    sigma.toDoubleVector(sigma_d);
    //debug_level_d = true;

    // exit gracefully
    //
    return true;
  }

  // method: computeLikelihood
  //
  // arguments:
  //  RVMTrainData& tdata: (input) training data storage
  //
  // return: double value with computed likelihood value from (24) of [1]
  //
  // this method computes the log likelihood of the weights given the
  // data. We would expect this value to increase as training proceeds
  //
  public double computeLikelihood() {
    // Debug
    // System.out.println("AlgorithmRVM : computeLikelihood()");

    // declare local variables
    //
    double result = 0;
    double divisor = phi_d.getNumColumns();

    // init the matrix
    Matrix weights_M = new Matrix();
    Matrix temp_M = new Matrix();
    Matrix result_M = new Matrix();
    weights_M.initMatrix(curr_weights_d);

    // compute the -1/2 * w * A * w' component of the likelihood
    //
    weights_M.multMatrix(A_d, temp_M);
    result_M.copyMatrix(weights_M);
    result_M.transposeMatrix(weights_M);

    temp_M.multMatrix(weights_M, result_M);
    result = result_M.getValue(0, 0);
    result *= 0.5;
    result /= divisor;

    // compute the summation in the log likelihood
    //  L = sum [log(sigma) * t + (1-t) * log(1-sigma)]
    //
    long len = sigma_d.size();
    double t;
    double sig;
    for (int i = 0; i < len; i++) {

      t = MathUtil.doubleValue(y_d, i);
      sig = MathUtil.doubleValue(sigma_d, i);

      // t == 0
      //
      if (t < 0.5) {
        result -= Math.log(1 - sig) / divisor;
      }

      // t == 1
      //
      else {
        result -= Math.log(sig) / divisor;
      }
    }

    // exit gracefully
    //
    return result;
  }

  // method: evaluateOutput
  //
  // arguments:
  //  const VectorDouble& point: (input) first example index
  //
  // return: result of the evaluation function
  //
  // this method computes the output of a specific example
  //
  double evaluateOutput(Vector point_a) {
    // Debug
    //// System.out.println("AlgorithmRVM : evaluateOutput(Vector point_a)");

    // declare local variables
    //
    double result = bias_d;

    // debug message
    //
    if (debug_level_d) {
      // System.out.println("bias_d: " + bias_d);
      // System.out.println("point_a: ");
      Matrix.printDoubleVector(point_a);
    }

    // evaluate the output at the example
    //
    for (int i = 0; i < weights_d.size(); i++) {
      double w_i = ( (Double) weights_d.get(i)).doubleValue();
      Vector x_i = (Vector) evalx_d.get(i);
      result += w_i * K(point_a, x_i);
      if (debug_level_d) {
        // System.out.println("x_i: " + x_i + "i: " + i + " result: " + result);
      }
    }

    // apply the logistic sigmoid link function
    //
    result = 1.0 / (1.0 + Math.exp( -result));

    // debug message
    //
    if (debug_level_d) {
      // System.out.println("result: " + result);
    }

    // return the result
    //
    return result;
  }

  // method: K
  //
  // arguments:
  //  Vector& point1: (input) first example
  //  Vector& point2: (input) second example
  //
  // return: Kernel evaluation
  //
  // this method evaluates the linear Kernel on the input vectors
  // K(x,y) = (x . y)
  //
  double K(Vector point1, Vector point2) {
    // Debug
    //// System.out.println("AlgorithmRVM : K(Vector point1, Vector point2)");

    double result = 0;

    if (kernel_type_d == KERNEL_TYPE_LINEAR) {
      result = MathUtil.linearKernel(point1, point2);
    }
    if (kernel_type_d == KERNEL_TYPE_RBF) {
      result = MathUtil.rbfKernel(point1, point2, 20);
    }
    if (kernel_type_d == KERNEL_TYPE_POLYNOMIAL) {
      result = MathUtil.polynomialKernel(point1, point2);
    }

    // return the result
    //
    return result;
  }

  // method: computeDecisionRegions
  //
  // arguments: none
  // return   : none
  //
  // method computes the line of discrimination for the classification
  // algorithms when the corresponding flags have been initialized
  //
  public void computeDecisionRegions() {
    // Debug
    // System.out.println("AlgorithmRVM : computeDecisionRegions()");

    DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();

    double currentX = scale.xmin;
    double currentY = scale.ymin;

    // set precision
    //
    int outputWidth = output_panel_d.disp_area_d.getXPrecision();
    int outputHeight = output_panel_d.disp_area_d.getYPrecision();

    double incrementY = (scale.ymax - scale.ymin) / outputHeight;
    double incrementX = (scale.xmax - scale.xmin) / outputWidth;

    // declare a 2D array to store the class associations
    //
    output_canvas_d = new int[outputWidth][outputHeight];

    // loop through each and every point on the pixmap and
    // determine which class each pixel is associated with
    //
    int associated = 0;

    pro_box_d.setProgressMin(0);
    pro_box_d.setProgressMax(outputWidth);

    pro_box_d.setProgressCurr(20);

    for (int i = 0; i < outputWidth; i++) {
      currentX += incrementX;
      currentY = scale.ymin;
      // set current status
      //
      pro_box_d.setProgressCurr(i);
      //pro_box_d.appendMessage(".");

      for (int j = 0; j < outputHeight; j++) {

        // declare the current pixel point
        //
        currentY += incrementY;
        MyPoint pixel = new MyPoint(currentX, currentY);

        Vector curr_point = new Vector();

        curr_point.add(new Double(pixel.x));
        curr_point.add(new Double(pixel.y));

        double output = evaluateOutput(curr_point);
        //// System.out.println(message.toString());
        if (output >= 0.5) {
          associated = 0;
          //decision_regions_d.add(pixel);
        }
        else {
          associated = 1;
        }

        // put and entry in the output canvas array to
        // indicate which class the current pixel is
        // closest to
        //
        output_canvas_d[i][j] = associated;

        // add a point to the vector of decision
        // region points if the class that the current
        // point is associated with is different for
        // the class what the previous point was
        // associated with i.e., a transition point
        //
        if (j > 0 && i > 0) {
          if (associated != output_canvas_d[i][j - 1] ||
              associated != output_canvas_d[i - 1][j]) {
            decision_regions_d.add(pixel);
          }
        }
      }
    } // end of the loop

  }

  //  // method computeErrors
  //  //
  //  // arguments:
  //  //    Data d: input data point
  //  //
  //  // return   : none
  //  //
  //  // display two matrices
  //  //
  public void computeErrors() {

    // declare local variables
    //
    String text;
    double error;
    int samples = 0;
    int samples1 = 0;
    int samples2 = 0;
    int samples3 = 0;
    int samples4 = 0;

    int incorrect = 0;
    int incorrect1 = 0;
    int incorrect2 = 0;
    int incorrect3 = 0;
    int incorrect4 = 0;

    DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
    // set scales
    int outputWidth = output_panel_d.disp_area_d.getXPrecision();
    int outputHeight = output_panel_d.disp_area_d.getYPrecision();

    double incrementY = (scale.ymax - scale.ymin) / outputHeight;
    double incrementX = (scale.xmax - scale.xmin) / outputWidth;

    // compute the classification error for the first set
    //
    for (int i = 0; i < set1_d.size(); i++) {

      MyPoint point = (MyPoint) set1_d.elementAt(i);
      samples1++;
      if ( (point.x > scale.xmin && point.x < scale.xmax)
          && (point.y > scale.ymin && point.y < scale.ymax)) {
        if (output_canvas_d[ (int) ( (point.x - scale.xmin) /
            incrementX)][ (int) ( (point.y - scale.ymin) / incrementY)] != 0) {
          incorrect1++;
        }
      }
    }

    if (set1_d.size() > 0) {

      // variable incorrect1 actually is number of correct
      //
      error = ( (double) incorrect1 / (double) samples1) * 100.0;

      error = 100.0 - error;

      text =
          new String(
          "       Results for class 0:\n"
          + "          Total number of samples: "
          + samples1
          + "\n"
          + "          Misclassified samples: "
          + (samples1 - incorrect1)
          + "\n"
          + "          Classification error: "
          + MathUtil.setDecimal(error, 2)
          + "%");

      pro_box_d.appendMessage(text);
    }

    // compute the classification error for the second set
    //
    for (int i = 0; i < set2_d.size(); i++) {

      MyPoint point = (MyPoint) set2_d.elementAt(i);
      samples2++;
      if ( (point.x > scale.xmin && point.x < scale.xmax)
          && (point.y > scale.ymin && point.y < scale.ymax)) {
        if (output_canvas_d[ (int) ( (point.x - scale.xmin) /
            incrementX)][ (int) ( (point.y - scale.ymin) / incrementY)] != 1) {
          incorrect2++;
        }
      }
    }

    if (set2_d.size() > 0) {

      // variable incorrect2 actually is number of correct
      //
      error = ( (double) incorrect2 / (double) samples2) * 100.0;

      error = 100.0 - error;

      text =
          new String(
          "       Results for class 1:\n"
          + "          Total number of samples: "
          + samples2
          + "\n"
          + "          Misclassified samples: "
          + (samples2 - incorrect2)
          + "\n"
          + "          Classification error: "
          + MathUtil.setDecimal(error, 2)
          + "%");

      pro_box_d.appendMessage(text);
    }

    // compute the overall classification error
    //
    samples = samples1 + samples2 + samples3 + samples4;
    incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4;

    // variable incorrect actually is number of correct
    //
    error = ( (double) incorrect / (double) samples) * 100.0;

    error = 100.0 - error;

    text =
        new String(
        "       Overall results:\n"
        + "          Total number of samples: "
        + samples
        + "\n"
        + "          Misclassified samples: "
        + (samples - incorrect)
        + "\n"
        + "          Classification error: "
        + MathUtil.setDecimal(error, 2)
        + "%");

    pro_box_d.appendMessage(text);

  }

}
