/**
 * file: MathUtil.java
 *
 * last edited: Sanjay Patil 01/24/2006
 * made changes in SetDecimal method
 */


import java.util.*;
import java.awt.*;

//These imports are not needed - Phil T. 6-23-03
//import javax.swing.*;

/**
 * class that round off floating point numbers to the specified number
 * of decimal places givem
 *
 */
public class MathUtil
{
    
    static final double DEF_BIAS = 0.0;
    static final double DEF_PENALTY = 50.0;
    static final double DEF_EPISLON = 1e-12;
    static final double DEF_TOLERANCE = 1e-3;
    static final double DEF_POLYNOMIAL_DEGREE = 3.0;
    static final double DEF_RBF_GAMMA = 0.5;
    static final double DEF_SIGMOID_KAPPA = 1.0;
    static final double DEF_SIGMOID_DELTA = 1.0;

    static final double MIN_DIFF_RATE = 0.00001;

    // declare the random number generator
    //
    public final static long RAND_SEED = 20022000;
    public final static Random random = new Random(RAND_SEED);

    // *********************************************************************
    //
    // class constructor
    //
    // *********************************************************************

    /**
     * Default constructor. Empty body
     */
    public MathUtil()
    {
	// default constructor
    }

    // *********************************************************************
    //
    // declare class methods
    //
    // *********************************************************************

    /**
     * method generates a random distribution centered about the mean with a
     * variance that is specified by the standard deviation.
     *
     * @param    mean mean of the distribution
     * @param    stddev standard deviation of the distribution
     * 
     * @return   random number centered about the mean 
     * 
     */
    public static double grand(double mean, double stddev)
    {

	// declare variables
	//
	double val = 0.0;
	double r1 = 0.0;
	double r0 = 0.0;

	// generate random number
	//
	r0 = random.nextDouble();
	r0 = r0 + 0.00000000001;
	r1 = random.nextDouble();

	val =
	    Math.sqrt(-2.0 * Math.log(r0)) * stddev * 
	               Math.cos(2 * Math.PI * r1) + mean;

	return val;
    }

    /**
     * method to round off floating point numbers to the specified number
     * of decimal places given
     *
     * @param   doubleNumber input number
     * @param   decimalPlaces number of decimal places to round
     *
     */
    static public double SetDecimal(double doubleNumber, int decimalPlaces)
    {

	long wholeNumber = (long)doubleNumber;
	double pastDecimal =
	    ((doubleNumber - wholeNumber) * Math.pow(10, decimalPlaces));
	long roundedDecimal = (long)Math.round(pastDecimal);
	long combine =
	    (long) (wholeNumber * Math.pow(10, decimalPlaces) 
		    + roundedDecimal);
	double result = combine / Math.pow(10, decimalPlaces);
	return result;    
    }

    /**
     * calculate the euclidean distance between the two points
     *
     * @param    x1 x-coordinate of the first point
     * @param    y1 y-coordinate of the first point
     * @param    x2 x-coordinate of the second point
     * @param    y2 y-coordinate of the second point
     * 
     * @return   distance between the points
     *
     */
    static public double distance(double x1, double y1, double x2, double y2)
    {
	
	double distance = 0.0;

	// claculate the euclidean distance
	//
	double deltaX = x2 - x1;
	double deltaY = y2 - y1;
	double sqrX = deltaX * deltaX;
	double sqrY = deltaY * deltaY;

	distance = Math.sqrt((double)sqrX + (double)sqrY);

	// return the distance
	//
	return distance;
    }

    /**
     * methods computes and returns the mean of the given cluster
     *
     * @param    vec Vector of points from the cluster
     * @return   MyPoint value of the cluster mean
     */
    static public MyPoint computeClusterMean(Vector vec)
    {
	
	// declare local variables
	//
	double xval = 0.0;
	double yval = 0.0;
	
	// compute the mean of the given cluster
	//
	for (int i = 0; i < vec.size(); i++)
	{
	    MyPoint point = (MyPoint)vec.elementAt(i);
	    xval += (double)point.x;
	    yval += (double)point.y;
	}

	xval = xval / vec.size();
	yval = yval / vec.size();
	
	return (new MyPoint(xval, yval));
    }

    /**
     * methods computes and returns the mean of the given cluster
     *
     * @param    vec Vector of points from the cluster
     * @return   MyPoint value of the mean
     *
     */
    static public MyPoint computePointMean(Vector vec)
    {

	// declare local variables
	//
	double xval = 0.0;
	double yval = 0.0;

	// compute the mean of the given cluster
	//
	for (int i = 0; i < vec.size(); i++)
        {
	    MyPoint point = (MyPoint)vec.elementAt(i);
	    xval += (double)point.x;
	    yval += (double)point.y;
	}

	xval = SetDecimal(xval / vec.size(), 3);
	yval = SetDecimal(yval / vec.size(), 3);
	
	return (new MyPoint((double)xval, (double)yval));
    }

    /**
     * methods computes and returns the mean of the given cluster
     *
     * @param   vec point from the cluster
     * @return  MyPoint value of the mean
     *
     */
    static public MyPoint computeMyPointMean(Vector vec)
    {

	// declare local variables
	//
	double xval = 0.0;
	double yval = 0.0;

	// compute the mean of the given cluster
	//
	for (int i = 0; i < vec.size(); i++)
	{
	    MyPoint point = (MyPoint)vec.elementAt(i);
	    xval += (double)point.x;
	    yval += (double)point.y;
	}

	xval = xval / vec.size();
	yval = yval / vec.size();
	
	return (new MyPoint(xval, yval));
    }
    
    /**
     * method takes in a decimal number and rounds to the given number
     * of decimal places passed
     *
     * @param   doubleNumber input number
     * @param   decimalPlaces number of significant decimal places 
     *
     * @return  double value of decimal with correct signifcant figures
     */
    public static double setDecimal(double doubleNumber, int decimalPlaces)
    {
	
	int wholeNumber = (int)doubleNumber;
	double pastDecimal =
	    ((doubleNumber - wholeNumber) * Math.pow(10, decimalPlaces));
	int roundedDecimal = (int)Math.round(pastDecimal);
	int combine =
	    (int) (wholeNumber * Math.pow(10, decimalPlaces) + roundedDecimal);
	double result = combine / Math.pow(10, decimalPlaces);
	
	// return the rounded decimal number
	//
	return result;
    }

    /**
     * this method evaluates the linear Kernel on the input vectors
     * K(x,y) = (x . y)
     *
     * @param    point1 First vector of points
     * @param    point2 Second vector of points
     *
     * @return  double value of Kernel evaluation
     *
     */
    public static double linearKernel(Vector point1, Vector point2)
    {

	// declare local variables
	//
	double result = 0.0;
	
	// if the length of the vectors are not equal error
	//
	
	/*
	  if (point1.length() != point2.length()) {
	  return Error::handle(name(), L"setPenalty", Error::ARG,
	  __FILE__, __LINE__);
	  }
	*/
	
	// compute the final result
	//
	result = vectorProduct(point1, point2);
	
	// return the result
	//
	return result;
    }
    
    /**
     * this method evaluates the redial basis function Kernel on the
     * input vectors with standard deviation sigma K(x,y) = exp(-gamma
     * * ((a.a)-2*(a.b)+(b.b)))
     *
     * @param    point1 First vector of points
     * @param    point2 Second vector of points
     *
     * @return double value of Kernel evaluation
     *
     */
    public static double rbfKernel(Vector point1, Vector point2)
    {
	
	// declare local variables
	//
	double result = 0.0;
	
	// compute the dot products for the rbf Kernel
	//
	double points1_sq = vectorProduct(point1, point1);
	double cross_prod = vectorProduct(point1, point2);
	double points2_sq = vectorProduct(point2, point2);
	
	// compute the final result
	//
	result =
	        Math.exp(-DEF_RBF_GAMMA * 
                          (points1_sq - 2 * cross_prod + points2_sq));
	
	// return the result
	//
	return result;
    }
    
    /**
     * this method evaluates the redial basis function Kernel on the
     * input vectors with standard deviation sigma K(x,y) = exp(-gamma
     * * ((a.a)-2*(a.b)+(b.b)))
     *
     * @param    point1 First vector of points
     * @param    point2 Second vector of points
     * @param    gamma_a Double value of gamma
     * @return double value of Kernel evaluation
     *
     */
    public static double rbfKernel(Vector point1, Vector point2, double gamma_a)
    {

	// declare local variables
	//
	double result = 0.0;

	// compute the dot products for the rbf Kernel
	//
	double points1_sq = vectorProduct(point1, point1);
	double cross_prod = vectorProduct(point1, point2);
	double points2_sq = vectorProduct(point2, point2);

	// compute the final result
	//
	result = Math.exp(
                          -gamma_a * 
                           (points1_sq - 2 * cross_prod + points2_sq));

	// return the result
	//
	return result;
    }

    /**
     * this method evaluates the second degree polynomial Kernel on the
     * input vectors K(x,y) = (x . y + 1)^p
     *
     * @param    points1 First vector of points
     * @param    points2 Second vector of points
     * @return  Double value of Kernel evaluation
     *
     */
    public static double polynomialKernel(Vector points1, Vector points2)
    {
	
	// declare local variables
	//
	double result = 0.0;
	
	// compute the cross product
	//
	double cross_prod = vectorProduct(points1, points2);
	
	// compute the final result
	//
	result = Math.pow(cross_prod + 1, DEF_POLYNOMIAL_DEGREE);
	
	// return the result
	//
	return result;
    }

    /**
     * this method evaluates the vector dot product
     *
     * @param    vec1 First vector of points
     * @param    vec2 Second vector of points
     *
     * @return  dot product
     *
     */
    public static double vectorProduct(Vector vec1, Vector vec2)
    {

	// declare local variables
	//
	double result = 0.0;

	// if the length of the vectors are not equal error
	//
	/*
	  if (vec1.size() != vec2.size()) {
	  return Error::handle(name(), L"setPenalty", Error::ARG,
	  __FILE__, __LINE__);
	  }
	*/

	// compute the dot product
	//
	for (int i = 0; i < vec1.size(); i++)
	{
	    double val1 = ((Double)vec1.get(i)).doubleValue();
	    double val2 = ((Double)vec2.get(i)).doubleValue();
	    result += (val1 * val2);
	}

	// return the result
	//
	return result;
    }

    /**
     * method sees if input vectors are "almost" equal
     *
     * @param    vec1 First vector of points
     * @param    vec2 Second vector of points
     * @return  True if all input vector components are "almost" equal
     */
    public static boolean almostEqual(Vector vec1, Vector vec2)
    {
	
	// compute the dot product
	//
	for (int i = 0; i < vec1.size(); i++)
	{
	    double val1 = ((Double)vec1.get(i)).doubleValue();
	    double val2 = ((Double)vec2.get(i)).doubleValue();
	    if (!almostEqual(val1, val2))
	    {
		return false;
	    }
	}
	
	// return the result
	//
	return true;
    }


    /**
     * method sees if input vectors are "almost" equal
     *
     * @param    vec1 First vector of points
     * @param    val2 Value to compare to
     * @return  True if all input vector components are "almost" equal
     */
    public static boolean almostEqual(Vector vec1, double val2)
    {

	// compute the dot product
	//
	for (int i = 0; i < vec1.size(); i++)
	{
	    double val1 = ((Double)vec1.get(i)).doubleValue();
	    if (!almostEqual(val1, val2))
            {
		return false;
	    }
	}
	
	// return the result
	//
	return true;
    }
    /**
     * method sees if input doubles are "almost" equal
     *
     * @param    val1 First double value
     * @param    val2 Second double value
     * @return  True if doubles are "almost" equal
     */
    public static boolean almostEqual(double val1, double val2)
    {
	
	// declare local variables
	//
	if (Math.abs(val1 - val2)
	     >= MIN_DIFF_RATE * (Math.abs(val1 + val2) + 0.01))
	{
	    return false;
	}
	
	// return the result
	//
	return true;
    }

    /**
     * initializes components of vector to double value passed
     *
     * @param   vec_a point from the cluster
     * @param   value_a value to set to vector
     *
     */
    static public void initDoubleVector(Vector<Double> vec_a, double value_a)
    {
	
	// loop over all elements
	//
	for (int i = 0; i < vec_a.size(); i++)
        {
	    vec_a.set(i, new Double(value_a));
	}
    }
    

    /**  
     * gets the double value of the input vector at the index passed
     *
     * @param    vec_a point from the cluster
     * @param    index_a
     * @return   double value of vector at indexed value
     *
     */
    static public double doubleValue(Vector<Double> vec_a, int index_a)
    {
	
	Double value = vec_a.get(index_a);
	return value.doubleValue();	
    }

    /**
     * Copies vector given specific parameters
     *
     * @param  vec1_a    vector to be overwritten/copyied to
     * @param  vec2_a    vector to be copied from
     * @param  length_a  number of elements to be copyied
     * @param  index1_a  index to start copying to first vector
     * @param  index2_a  index to start copying from second vector
     *
     */
    static public void copyVector( Vector<Double> vec1_a,  Vector<Double> vec2_a, int length_a, int index1_a, int index2_a)
    {
	
	// loop over all elements
	//
	for (int i = 0; i < length_a; i++)
	{
	    vec1_a.set(i + index1_a, vec2_a.get(i + index2_a));
	}
	
    }
    
    /**
     * Within Class
     *
     * @param    data_a input data points
     * @param    rx_a double input of x
     * @param    ry_a double input of y
     *
     */
    public static void withinClass(DataPoints data_a, double rx_a, double ry_a)
    {
	
	// declare local variables
	//
	int size = 0;
	double x[] = null;
	double y[] = null;
	
	// declare the covariance object
	//
	Covariance cov = new Covariance();
	
	// declare local matrices
	//
	Matrix M = new Matrix();
	Matrix M1 = new Matrix();
	Matrix M2 = new Matrix();
	Matrix M3 = new Matrix();
	Matrix M4 = new Matrix();
	
	// set scales
	
	double Xmin = OutputPanel.Xmin;
	double Ymax = OutputPanel.Ymax;
	
	// get the Rx and Ry ratio, seconds per pixel
	//
	double Rx = rx_a;
	double Ry = ry_a;
    
	// get the first data set size
	//
	size = data_a.dset1.size();

	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];

	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    
	    MyPoint p = (MyPoint)data_a.dset1.elementAt(i);
	    x[i] = (p.x * Rx) + Xmin;
	    y[i] = Ymax - (p.y * Ry);
	}
	
	// compute the covariance matrix of the first data set
	//
	M1.row = M1.col = 2;
	M1.Elem = new double[2][2];
	M1.resetMatrix();

	if (size > 0)
	{
	    M1.Elem = cov.computeCovariance(x, y);
	}

	// get the second data set size
	//
	size = data_a.dset2.size();

	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];

	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    
	    MyPoint p = (MyPoint)data_a.dset2.elementAt(i);
	    x[i] = (p.x * Rx) + Xmin;
	    y[i] = Ymax - (p.y * Ry);
	}
	
	// compute the covariance matrix of the second data set
	//
	M2.row = M2.col = 2;
	M2.Elem = new double[2][2];
	M2.resetMatrix();

	if (size > 0)
	{
	    M2.Elem = cov.computeCovariance(x, y);
	}

	// get the third data set size
	//
	size = data_a.dset3.size();
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    
	    MyPoint p = (MyPoint)data_a.dset3.elementAt(i);
	    x[i] = (p.x * Rx) + Xmin;
	    y[i] = Ymax - (p.y * Ry);
	}

	// compute the covariance matrix of the third data set
	//
	M3.row = M3.col = 2;
	M3.Elem = new double[2][2];
	M3.resetMatrix();
	
	if (size > 0)
	{
	    M3.Elem = cov.computeCovariance(x, y);
	}
	
	// get the fourth data set size
	//
	size = data_a.dset4.size();
	
	// initialize arrays to store the samples
	//
	x = new double[size];
	y = new double[size];

	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
	{
	    
	    MyPoint p = (MyPoint)data_a.dset4.elementAt(i);
	    x[i] = (p.x * Rx) + Xmin;
	    y[i] = Ymax - (p.y * Ry);
	}

	// compute the covariance matrix of the fourth data set
	//
	M4.row = M4.col = 2;
	M4.Elem = new double[2][2];
	M4.resetMatrix();

	if (size > 0)
        {
	    M4.Elem = cov.computeCovariance(x, y);
	}

	// compute the within class scatter matrix
	//
	M.row = M.col = 2;
	M.Elem = new double[2][2];
	M.resetMatrix();
	M.addMatrix(M1);
	M.addMatrix(M2);
	M.addMatrix(M3);
	M.addMatrix(M4);
    }

    /**
     * this method determines the between class scatter matrix for
     * the class independent linear discrimination algorithm
     *
     * @param    d input data points
     * @param    M between class scatter matrix
     * @param    rx_a double input of x
     * @param    ry_a double input of y
     */
    public static void betweenClass(
				    DataPoints d, Matrix M,
				    double rx_a, double ry_a)
    {

	// declare local variables
	//
	double xmean = 0.0;
	double ymean = 0.0;

	double xmean1 = 0.0;
	double ymean1 = 0.0;
	double xmean2 = 0.0;
	double ymean2 = 0.0;
	double xmean3 = 0.0;
	double ymean3 = 0.0;
	double xmean4 = 0.0;
	double ymean4 = 0.0;

	// declare local matrices
	//
	Matrix M1 = new Matrix();
	Matrix T1 = new Matrix();
	Matrix M2 = new Matrix();
	Matrix T2 = new Matrix();
	Matrix M3 = new Matrix();
	Matrix T3 = new Matrix();
	Matrix M4 = new Matrix();
	Matrix T4 = new Matrix();

	// declare the initial random variables
	//
	double transpose[][] = new double[2][1];
	double mean[][] = new double[1][2];

	// set scales
	//
	double Xmin = OutputPanel.Xmin;
	double Ymax = OutputPanel.Ymax;

	// get the Rx and Ry ratio, seconds per pixel
	//
	double Rx = rx_a;
	double Ry = ry_a;

	// initialize the between class matrix
	//
	M.row = M.col = 2;
	M.Elem = new double[2][2];
	M.resetMatrix();

	// obtain the means of each individual class
	//
	if (d.dset1.size() > 0)
        {
	    MyPoint p = (MyPoint)computePointMean(d.dset1);
	    xmean1 = (p.x * Rx) + Xmin;
	    ymean1 = Ymax - (p.y * Ry);
	}

	if (d.dset2.size() > 0)
        {
	    MyPoint p = (MyPoint)computePointMean(d.dset2);
	    xmean2 = (p.x * Rx) + Xmin;
	    ymean2 = Ymax - (p.y * Ry);
	}

	if (d.dset3.size() > 0)
        {
	    MyPoint p = (MyPoint)computePointMean(d.dset3);
	    xmean3 = (p.x * Rx) + Xmin;
	    ymean3 = Ymax - (p.y * Ry);
	}

	if (d.dset4.size() > 0)
	{
	    MyPoint p = (MyPoint)computePointMean(d.dset4);
	    xmean4 = (p.x * Rx) + Xmin;
	    ymean4 = Ymax - (p.y * Ry);
	}

	// compute the global mean of all data sets
	//
	int samples = 0;
	
	for (int i = 0; i < d.dset1.size(); samples++, i++)
	{
	    MyPoint p2 = (MyPoint)d.dset1.elementAt(i);
	    xmean += (p2.x * Rx) + Xmin;
	    ymean += Ymax - (p2.y * Ry);
	}

	for (int i = 0; i < d.dset2.size(); samples++, i++)
	{
	    MyPoint p2 = (MyPoint)d.dset2.elementAt(i);
	    xmean += (p2.x * Rx) + Xmin;
	    ymean += Ymax - (p2.y * Ry);
	}

	for (int i = 0; i < d.dset3.size(); samples++, i++)
	{
	    MyPoint p2 = (MyPoint)d.dset3.elementAt(i);
	    xmean += (p2.x * Rx) + Xmin;
	    ymean += Ymax - (p2.y * Ry);
	}

	for (int i = 0; i < d.dset4.size(); samples++, i++)
	{
	    MyPoint p2 = (MyPoint)d.dset4.elementAt(i);
	    xmean += (p2.x * Rx) + Xmin;
	    ymean += Ymax - (p2.y * Ry);
	}
	
	xmean /= samples;
	ymean /= samples;
	
	// compute the between class scatter contribution of the first set
	//
	if (d.dset1.size() > 0)
        {

	    Matrix S = new Matrix();

	    mean[0][0] = xmean1 - xmean;
	    mean[0][1] = ymean1 - ymean;
	    M1.initMatrix(mean, 1, 2);

	    transpose[0][0] = xmean1 - xmean;
	    transpose[1][0] = ymean1 - ymean;
	    T1.initMatrix(transpose, 2, 1);

	    T1.multMatrix(M1, S);
	    M.addMatrix(S);
	}

	// compute the between class scatter contribution of the second set
	//
	if (d.dset2.size() > 0)
	{

	    Matrix S = new Matrix();

	    mean[0][0] = xmean2 - xmean;
	    mean[0][1] = ymean2 - ymean;
	    M2.initMatrix(mean, 1, 2);

	    transpose[0][0] = xmean2 - xmean;
	    transpose[1][0] = ymean2 - ymean;
	    T2.initMatrix(transpose, 2, 1);

	    T2.multMatrix(M2, S);
	    M.addMatrix(S);
	}

	// compute the between class scatter contribution of the third set
	//
	if (d.dset3.size() > 0)
       {

	   Matrix S = new Matrix();

	   mean[0][0] = xmean3 - xmean;
	   mean[0][1] = ymean3 - ymean;
	   M3.initMatrix(mean, 1, 2);

	   transpose[0][0] = xmean3 - xmean;
	   transpose[1][0] = ymean3 - ymean;
	   T3.initMatrix(transpose, 2, 1);

	   T3.multMatrix(M3, S);
	   M.addMatrix(S);
       }

	// compute the between class scatter contribution of the forth set
	//
	if (d.dset4.size() > 0)
        {

	    Matrix S = new Matrix();

	    mean[0][0] = xmean4 - xmean;
	    mean[0][1] = ymean4 - ymean;
	    M4.initMatrix(mean, 1, 2);

	    transpose[0][0] = xmean4 - xmean;
	    transpose[1][0] = ymean4 - ymean;
	    T4.initMatrix(transpose, 2, 1);

	    T4.multMatrix(M4, S);
	    M.addMatrix(S);
	}
    }

}
