/**
 * @(#) AlgorithmPCA.java   v6.0 05/23/2005
 * last edited Ryan Irwin 
 * document created  02/09/03
 *
 * Defines the operation of the PCA Class-Independent Algorithm
 *
 */


// import java packages
//
import java.awt.*;
import java.util.*;

/**
 * Operation of the PCA Class-Independent Algorithm
 */
public class AlgorithmPCA extends Algorithm
{
    
    //-----------------------------------------------------------------
    //
    // instance data members
    //
    //-----------------------------------------------------------------
    
    // vector of support region points
    //
    Matrix trans_matrix_d = new Matrix();
    Matrix cov_matrix_d = new Matrix();
    Vector<MyPoint> support_vectors_d = new Vector<MyPoint>();
    Vector<MyPoint> decision_regions_d = new Vector<MyPoint>();
    int output_canvas_d[][];
    

    //-----------------------------------------------------------------
    //
    // public class methods
    //
    //------------------------------------------------------------------

    /**
     * Overrides the initialize() method in the base class.  Initializes
     * member data and prepares for execution of first step.  This method
     * "resets" the algorithm.
     *
     * @return true
     */
    public boolean initialize()
    {
	// Debug
	//
	// System.out.println(algo_id + ": initialize()");
	
	trans_matrix_d = new Matrix();
	cov_matrix_d = new Matrix();
	support_vectors_d = new Vector<MyPoint>();
	point_means_d = new Vector<MyPoint>();
	decision_regions_d = new Vector<MyPoint>();
	step_count = 3;
	algo_id = "AlgorithmPCA";
	
	// add the process description for the PCA algorithm
	//
	if (description_d.size() == 0)
        {
	    String str = new String("   0. Initialize the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   1. Displaying the original data.");
	    description_d.addElement(str);
	    
	    str = new String("   2. Computing the support regions.");
	    description_d.addElement(str);
	    
	    str = new String("   3. Computing the decision regions based on the class independent principal component analysis algorithm.");
	    description_d.addElement(str);
	}
	    
	// append message to process box
	//
	pro_box_d.appendMessage("Class Independent Principal" 
				+ " Component Analysis:" + "\n");

	// set the data points for this algorithm
	//
	//	set1_d = (Vector)data_points_d.dset1.clone();
	//	set2_d = (Vector)data_points_d.dset2.clone();
	//	set3_d = (Vector)data_points_d.dset3.clone();
	//	set4_d = (Vector)data_points_d.dset4.clone();
	//
	set1_d = data_points_d.dset1;
	set2_d = data_points_d.dset2;
	set3_d = data_points_d.dset3;
	set4_d = data_points_d.dset4;
	
	// set the step index
	//
	step_index_d = 0;
	
	// append message to process box
	//
	pro_box_d.appendMessage((String)description_d.get(step_index_d));
	
	// exit gracefully
	//
	return true;
    }
    
    /**
     * Implementation of the run function from the Runnable interface.
     * Determines what the current step is and calls the appropriate method.
     */
    public void run()
    {
	// Debug
	//
	// System.out.println(algo_id + ": run()");
	
	if (step_index_d == 1)
	{
	    disableControl();
	    step1();
	    enableControl();
	}
	    
        else if (step_index_d == 2)
	{
	    disableControl();
	    step2(); 
	    enableControl();
	}
	
	else if (step_index_d == 3)
        {
	    disableControl();
	    step3();
	    pro_box_d.appendMessage("   Algorithm Complete");
	    enableControl(); 
	}
	
	// exit gracefully
	//
	return;
    }
    
    /**
     *
     * step one of the algorithm. Scales the display to fit the plot.
     *
     * @return true
     */
    boolean step1()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step1()");
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(1);
	pro_box_d.setProgressCurr(0);
	
	scaleToFitData();
	
	// Display original data
	//
	output_panel_d.addOutput(set1_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset1);
	output_panel_d.addOutput(set2_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset2);
	output_panel_d.addOutput(set3_d, Classify.PTYPE_INPUT,
				 data_points_d.color_dset3);
	output_panel_d.addOutput(set4_d, Classify.PTYPE_INPUT, 
				 data_points_d.color_dset4);
	
	// step 1 completed
	//
	pro_box_d.setProgressCurr(1);
	output_panel_d.repaint();
	
	// exit gracefully
	//
	return true;
    }

    /**
     *
     * step two of the algorithm. Finds the PCA for the given data
     *
     * @return true
     */
    boolean step2()
    {
	// Debug
	//System.out.println(algo_id + ": step2()");
	
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
	
	// append message to process box
	//
	transformPCA();
	printMatrices();
	computeMeans();
	
	// display means
	//
	output_panel_d.addOutput(point_means_d, Classify.PTYPE_OUTPUT_LARGE, 
				 Color.black);
	
	// display support vectors
	//
	output_panel_d.addOutput(support_vectors_d, Classify.PTYPE_INPUT, 
				 Color.cyan);
	
	// display support vectors
	//
	pro_box_d.setProgressCurr(20);
	output_panel_d.repaint();
	
	// exit gracefully
	//
	return true;
    }

    /**
     *
     * step three of the algorithm. Computes the decision regions
     *
     * @return true
     */
    boolean step3()
    {
	// Debug
	//
	// System.out.println(algo_id + ": step3()");
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(20);
	pro_box_d.setProgressCurr(0);
	
	// compute the decision regisions
	//
	computeDecisionRegions();
	    
	// compute errors
	//
	computeErrors();
	
	// display support vectors
	//
	// display support vectors
	//
	output_panel_d.addOutput(decision_regions_d, Classify.PTYPE_INPUT, 
				 new Color(255, 200, 0));
	
	output_panel_d.repaint();
	
	// exit gracefully
	//
	return true;
    }
    
    /**
     * 
     * transforms a given set of points to a new space
     * using the class independent principal component analysis algorithm
     *
     */
    public void transformPCA()
    {
	// Debug
	// 
	// System.out.println(algo_id + ": transformPCA()");
	    
	// declare local variables
	//
	int size = 0;
	int xsize = 0;
	int ysize = 0;
	    
	// declare variables to compute the global mean
	//
	double xval = 0.0;
	double yval = 0.0;
	double xmean = 0.0;
	double ymean = 0.0;
	    
	// declare the covariance object
	//
	Covariance cov = new Covariance();
	    
	// declare an eigen object
	//
	// Since Eigen is a class of static member functions 
	// it is not correct to instantiate it - Phil T. 6-23-03 
	//Eigen eigen = new Eigen();
	
	// declare the covariance matrix
	//
	Matrix covariance = new Matrix();
	covariance.row = covariance.col = 2;
	covariance.Elem = new double[2][2];
	    
	// declare arrays for the eigenvalues
	//
	double eigVal[] = null;
	    
	// declare an array to store the eigen vectors
	//
	double eigVec[] = new double[2];
	    
	// declare matrix objects
	//
	Matrix T = new Matrix();
	Matrix M = new Matrix();
	Matrix W = new Matrix();
	    
	// allocate memory for the matrix elements
	//
	T.Elem = new double[2][2];
	M.Elem = new double[2][2];
	W.Elem = new double[2][2];
	    
	// declare arrays to store the samples
	//
	double x[] = null;
	double y[] = null;
	    
	// declare the maximum size of all data sets together
	//
	int maxsize = set1_d.size() + set2_d.size() 
	            + set3_d.size() + set4_d.size();
	    
	// initialize arrays to store the samples
	//
	x = new double[maxsize];
	y = new double[maxsize];
	    
	// get the samples from the first data set
	//
	size = set1_d.size();
	    
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
        {
	    MyPoint p = (MyPoint)set1_d.elementAt(i);
	    xval += p.x;
	    yval += p.y;
	    x[xsize++] = p.x;
	    y[ysize++] = p.y;
	}
	
	// get the samples from the second data set
	//
	size = set2_d.size();
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
        {
	    MyPoint p = (MyPoint)set2_d.elementAt(i);
	    xval += p.x;
	    yval += p.y;
	    x[xsize++] = p.x;
	    y[ysize++] = p.y;
	}
	
	// get the samples from the third data set
	//
	size = set3_d.size();
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
        {
	    MyPoint p = (MyPoint)set3_d.elementAt(i);
	    xval += p.x;
	    yval += p.y;
	    x[xsize++] = p.x;
	    y[ysize++] = p.y;
	}
	
	// get the samples from the first data set
	//
	size = set4_d.size();
	
	// set up the initial random vectors i.e., the vectors of
	// X and Y coordinate points form the display
	//
	for (int i = 0; i < size; i++)
        {
	    MyPoint p = (MyPoint)set4_d.elementAt(i);
	    xval += p.x;
	    yval += p.y;
	    x[xsize++] = p.x;
	    y[ysize++] = p.y;
	}
	
	if (maxsize > 0)
	{
	    
	    // initialize the transformation matrix dimensions
	    //
	    W.row = 2;
	    W.col = 2;
	    
	    // reset the matrices
	    //
	    W.resetMatrix();
	    
	    // compute the covariance matrix of the first data set
	    //
	    covariance.Elem = cov.computeCovariance(x, y);
	    cov_matrix_d = covariance;
	    
	    // initialize the matrix needed to compute the eigenvalues
	    //
	    T.initMatrix(covariance.Elem, 2, 2);
	    
	    // make a copy of the original matrix
	    //
	    M.copyMatrix(T);
	    
	    // compute the eigen values
	    //
	    // Changed eigen to Eigen since member function is static -
	    // Phil T. 6-23-03
	    eigVal = Eigen.compEigenVal(T);
	    
	    // compute the eigen vectors
	    //
	    for (int i = 0; i < 2; i++)
	    {
		// Changed eigen to Eigen since member function is static - 
		// Phil T. 6-23-03
		Eigen.calcEigVec(M, eigVal[i], eigVec);
		for (int j = 0; j < 2; j++)
		{
		    W.Elem[j][i] = eigVec[j] / Math.sqrt(eigVal[i]);
		}
	    }
	    
	    // save the transformation matrix 
	    //
	    trans_matrix_d = W;
	}
	
	// compute the global mean of the data sets
	//
	xmean = xval / xsize;
	ymean = yval / ysize;
	
	// determine points for the support regions
	//
	double val[][] = new double[2][1];
	
	Matrix invT = new Matrix();
	Matrix supp = new Matrix();
	Matrix temp = new Matrix();
	
	// set up the angle with which to rotate the axis
	//
	double theta = 0.0;
	double alpha = cov_matrix_d.Elem[0][0] - cov_matrix_d.Elem[1][1];
	double beta = -2 * cov_matrix_d.Elem[0][1];
	    
	if (eigVal[0] > eigVal[1])
	{
	    theta = Math.atan2((alpha - Math.sqrt((alpha * alpha) 
						  + (beta * beta))), beta);
	}
	else
	{
	    theta = Math.atan2((alpha + Math.sqrt((alpha * alpha) 
						  + (beta * beta))), beta);
	}
	
	// compute the inverse of the transformation matrix
	//
	trans_matrix_d.invertMatrix(invT);
	    
	// loop through all points on the circumference of the 
	// gaussian sphere in the transformed space
	//
	for (int i = 0; i < 360; i++)
        {
		
	    // get the x and y co-ordinates
	    //
	    val[0][0] = 1.5 * Math.cos(i);
	    val[1][0] = 1.5 * Math.sin(i);
	    
	    // set up the points as a matrix in order for multiplication
	    //
	    temp.initMatrix(val, 2, 1);
	    
	    // transform the points from the feature space back to the
	    // original space to create the support region for the data set
	    //
	    invT.multMatrix(temp, supp);
	    
	    // rotate the points after transforming them to the new space
	    //
	    xval = (supp.Elem[0][0] * Math.cos(theta)) 
		 - (supp.Elem[1][0] * Math.sin(theta));
	    yval = (supp.Elem[0][0] * Math.sin(theta)) 
		 + (supp.Elem[1][0] * Math.cos(theta));
	    
	    // time shift the co-ordinates to the global mean
	    //
	    xval = xval + xmean;
	    yval = yval + ymean;
	    
	    // add the point to the support region vector
	    //
	    MyPoint pt = new MyPoint(xval, yval);
	    support_vectors_d.addElement(pt);
	}
    }

    /**
     * Computes the line of discrimination for the classification 
     * algorithms when the corresponding flags have been initialized
     *
     */
    public void computeDecisionRegions()
    {
	// Debug
	//
	// System.out.println(algo_id + ": computeDecisionRegions()");
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale();
	
	double currentX = scale.xmin;
	double currentY = scale.ymin;
	
	// set precision
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	
	double incrementY = (scale.ymax - scale.ymin) / outputHeight;
	double incrementX = (scale.xmax - scale.xmin) / outputWidth;
	
	// declare a 2D array to store the class associations
	//
	output_canvas_d = new int[outputWidth][outputHeight];
	
	// loop through each and every point on the pixmap and
	// determine which class each pixel is associated with
	//
	MyPoint point;
	double dist = 0.0;
	int associated = 0;
	double smallestSoFar = Double.MAX_VALUE;
	
	int target = 0;
	boolean set1flag = true;
	boolean set2flag = true;
	boolean set3flag = true;
	
	pro_box_d.setProgressMin(0);
	pro_box_d.setProgressMax(outputWidth);
	pro_box_d.setProgressCurr(0);
	
	for (int i = 0; i < outputWidth; i++)
	{
	    currentX += incrementX;
	    currentY = scale.ymin;
	    pro_box_d.setProgressCurr(i);    
	    
	    for (int j = 0; j < outputHeight; j++)
	    {
		
		// declare the current pixel point
		//
		currentY += incrementY;
		MyPoint pixel = new MyPoint(currentX, currentY);
		smallestSoFar = Double.MAX_VALUE;
		
		// convert the pixel to the time domain
		//
		double X[][] = new double[1][2];
		X[0][0] = pixel.x;
		X[0][1] = pixel.y;
		
		// reset the boolean flags
		//
		set1flag = true;
		set2flag = true;
		set3flag = true;
		
		// find the closest point from the first class
		//
		for (int k = 0; k < point_means_d.size(); k++)
	        {
		    
		    // classify the sample to the first set
		    //
		    if (set1_d.size() > 0 && set1flag)
		    {
			set1flag = false;
			target = 0;
		    }
		    
		    // classify the sample to the second set
		    //
		    else if (set2_d.size() > 0 && set2flag)
		    {
			set2flag = false;
			target = 1;
		    }

		    // classify the sample to the third set
		    //
		    else if (set3_d.size() > 0 && set3flag)
		    {
			set3flag = false;
			target = 2;
		    }
		    
		    // classify the sample to the forth set
		    //
		    else
		    {
			target = 3;
		    }
		    
		    // get the first mean point
		    //
		    point = (MyPoint)point_means_d.elementAt(k);
		    
		    // convert the mean point to the time domain
		    //
		    double Y[][] = new double[1][2];
		    Y[0][0] = point.x;
		    Y[0][1] = point.y;
		    
		    // represent the pixel as a matrix
		    //
		    Matrix A = new Matrix();
		    A.initMatrix(X, 1, 2);
		    
		    // represent the mean point as a matrix
		    //
		    Matrix B = new Matrix();
		    B.initMatrix(Y, 1, 2);
		    
		    // transform the pixel and mean point to the 
		    // feature space
		    //
		    Matrix C = new Matrix();
		    Matrix D = new Matrix();
		    
		    A.multMatrix(trans_matrix_d, C);
		    B.multMatrix(trans_matrix_d, D);
		    
		    // find the distance between the pixel and
		    // mean point
		    //
		    dist = MathUtil.distance(C.Elem[0][0], C.Elem[0][1], 
					     D.Elem[0][0], D.Elem[0][1]);
			
		    if (dist < smallestSoFar)
		    {
			
			associated = target;
			smallestSoFar = dist;
		    }
		}
		
		// put and entry in the output canvas array to
		// indicate which class the current pixel is
		// closest to
		//
		output_canvas_d[i][j] = associated;
		
		// add a point to the vector of decision
		// region points if the class that the current
		// point is associated with is different for
		// the class what the previous point was
		// associated with i.e., a transition point
		//
		if (j > 0 && i > 0)
		{
		    if (associated != output_canvas_d[i][j - 1] 
			|| associated != output_canvas_d[i - 1][j])
		    {
			decision_regions_d.add(pixel);
		    }
		}
	    }
	} // end of the loop
	
    }
    
    /**
     *
     * Computes the number of data points in classification error 
     *
     */
    public void computeErrors()
    {
	
	// declare local variables
	//
	String text;
	double error;
	int samples = 0;
	int samples1 = 0;
	int samples2 = 0;
	int samples3 = 0;
	int samples4 = 0;
	
	int incorrect = 0;
	int incorrect1 = 0;
	int incorrect2 = 0;
	int incorrect3 = 0;
	int incorrect4 = 0;
	
	DisplayScale scale = output_panel_d.disp_area_d.getDisplayScale(); 

	// set scales
	//
	int outputWidth = output_panel_d.disp_area_d.getXPrecision();
	int outputHeight = output_panel_d.disp_area_d.getYPrecision();
	
	double incrementY = (scale.ymax - scale.ymin) / outputHeight;
	double incrementX = (scale.xmax - scale.xmin) / outputWidth;
	
	// compute the classification error for the first set
	//
	for (int i = 0; i < set1_d.size(); i++)
        {
	    MyPoint point = (MyPoint)set1_d.elementAt(i);
	    samples1++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 0)
		{
		    incorrect1++;
		}
	    }
	}
	
	if (set1_d.size() > 0)
        {
	    
	    error = ((double)incorrect1 / (double)samples1) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 0:\n"
			   + "          Total number of samples: "
			   + samples1
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect1
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the second set
	//
	for (int i = 0; i < set2_d.size(); i++)
        {
	    
	    MyPoint point = (MyPoint)set2_d.elementAt(i);
	    samples2++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 1)
		{
		    incorrect2++;
		}
	    }
	}
	
	if (set2_d.size() > 0)
        {
	    
	    error = ((double)incorrect2 / (double)samples2) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 1:\n"
			   + "          Total number of samples: "
			   + samples2
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect2
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the third set
	//
	for (int i = 0; i < set3_d.size(); i++)
        {
	    
	    MyPoint point = (MyPoint)set3_d.elementAt(i);
	    samples3++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 2)
		{
		    incorrect3++;
		}
	    }
	}
	
	if (set3_d.size() > 0)
        {
		
	    error = ((double)incorrect3 / (double)samples3) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 2:\n"
			   + "          Total number of samples: "
			   + samples3
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect3
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the classification error for the forth set
	//
	for (int i = 0; i < set4_d.size(); i++)
        {
		
	    MyPoint point = (MyPoint)set4_d.elementAt(i);
	    samples4++;
	    if ((point.x > scale.xmin && point.x < scale.xmax)
		&& (point.y > scale.ymin && point.y < scale.ymax))
	    {
		if (output_canvas_d[(int)((point.x - scale.xmin) / incrementX)]
		    [(int)((point.y - scale.ymin) / incrementY)] != 3)
		{
		    incorrect4++;
		}
	    }
	}
	
	if (set4_d.size() > 0)
        {
	    
	    error = ((double)incorrect4 / (double)samples4) * 100.0;
	    
	    text =
		new String(
			   "       Results for class 3:\n"
			   + "          Total number of samples: "
			   + samples4
			   + "\n"
			   + "          Misclassified samples: "
			   + incorrect4
			   + "\n"
			   + "          Classification error: "
			   + MathUtil.setDecimal(error, 2)
			   + "%");
	    
	    pro_box_d.appendMessage(text);
	}
	
	// compute the overall classification error
	//
	samples = samples1 + samples2 + samples3 + samples4;
	incorrect = incorrect1 + incorrect2 + incorrect3 + incorrect4;
	error = ((double)incorrect / (double)samples) * 100.0;
	
	text =
	    new String(
		       "       Overall results:\n"
		       + "          Total number of samples: "
		       + samples
		       + "\n"
		       + "          Misclassified samples: "
		       + incorrect
		       + "\n"
		       + "          Classification error: "
		       + MathUtil.setDecimal(error, 2)
		       + "%");
	
	pro_box_d.appendMessage(text);
	
    }    

    /**
     * Appends messages to the pro_box_d variable
     */
    public void printMatrices()
    {
	double a11, a12, a21, a22;
	String text;

	a11 = MathUtil.setDecimal(cov_matrix_d.Elem[0][0], 2);
	a12 = MathUtil.setDecimal(cov_matrix_d.Elem[0][1], 2);
	a21 = MathUtil.setDecimal(cov_matrix_d.Elem[1][0], 2);
	a22 = MathUtil.setDecimal(cov_matrix_d.Elem[1][1], 2);
	
	text = new String("      Covariance matrix:\n" +
			  "         " + a11 + "    " + a12 + "\n" +
			  "         " + a21 + "    " + a22);
	
	pro_box_d.appendMessage(text + "\n");
	
	a11 = MathUtil.setDecimal(trans_matrix_d.Elem[0][0], 2);
	a12 = MathUtil.setDecimal(trans_matrix_d.Elem[0][1], 2);
	a21 = MathUtil.setDecimal(trans_matrix_d.Elem[1][0], 2);
	a22 = MathUtil.setDecimal(trans_matrix_d.Elem[1][1], 2);
	
	text = new String("      Transformation matrix:\n" +
			  "         " + a11 + "    " + a12 + "\n" +
			  "         " + a21 + "    " + a22);
	
	pro_box_d.appendMessage(text + "\n");
    }
}

