/* file: DataPoints.java
 *
 *  last edited: Ryan Irwin
 *
 */

// import necessary java libraries
//
import java.awt.*;
import java.util.*;
import javax.swing.*;

//These imports are not needed - Phil T. 6-23-03
//import javax.swing.*;
//import java.awt.event.*;
//import java.applet.Applet;

/**
 * class holds the input DataPoints classes that are to be classified as well
 * as the classification algorithms needed to compute the decision regions
 *
 */
public class DataPoints
{
    
    // *********************************************************************
    //
    // declare global variables and components
    //
    // *********************************************************************
    
    static final int DTYPE_USER_SELECTED = 0;   
    
    // declare vectors needed to store the set points
    //
    int set_index_d = 1;
    int type_d = DTYPE_USER_SELECTED;
    int selectFlag = 0;
    int selStartX = 0;
    int selStartY = 0;
    int selEndX = 0;
    int selEndY = 0;

    // the set with double value
    //
    Vector<MyPoint> dset1;
    Vector<MyPoint> dset2;
    Vector<MyPoint> dset3;
    Vector<MyPoint> dset4;

    // colors for each set
    //
    Color color_dset1 = Color.red;  
    Color color_dset2 = Color.green;
    Color color_dset3 = Color.blue;
    Color color_dset4 = Color.orange;

    // gaussian setup
    //  
    int points = 25;
    double cov11 = 0.05;
    double cov12 = 0;
    double cov21 = 0;
    double cov22 = 0.05;

    // Scale values
    //

    // double xmin = -10.0;
    // double xmax = 10.0;
    // double ymin = -10.0;
    // double ymax = 10.0;  
 
    /**
     * initializes the samples to be plotted in the signal panel by calling
     * initialize()
     *
     */
    DataPoints()
    {
	
	// initialize the data set 
	//
	initialize();
	
    }
    
    // *********************************************************************
    //
    // declare class methods
    //
    // *********************************************************************
    
    /**
     * initializes dset1-dset4
     *
     * @return   true
     */
    public boolean initialize()
    {
	
	// initialize the data set 
	//
	dset1 = new Vector<MyPoint>(40, 20);
	dset2 = new Vector<MyPoint>(40, 20);
	dset3 = new Vector<MyPoint>(40, 20);
	dset4 = new Vector<MyPoint>(40, 20);
	
	return true;
    }
    
    /**
     * method returns true if valid data is present else it returns false
     *
     * @return     true if data is valid
     */
    public boolean isDataValid()
    {

	// check if valid data is present in the first data set
	//
	if (dset1.size() > 0)
	{
	    return true;
	}
	
	// check if valid data is present in the second data set
	//
	if (dset2.size() > 0)
	{
	    return true;
	}

	// check if valid data is present in the third data set
	//
	if (dset3.size() > 0)
	{
	    return true;
	}
	
	// check if valid data is present in the fourth data set
	//
	if (dset4.size() > 0)
	{
	    return true;
	}
	
	// indicate that no valid data is present
	//
	return false;
    }
    
    /**
     * clear out all point form the data sets
     *
     */
    public void clearAllSets()
    {
	
	if (dset1.size() > 0)
	{
	    dset1.removeAllElements();
	}
	if (dset2.size() > 0)
	{
	    dset2.removeAllElements();
	}
	if (dset3.size() > 0)
	{
	    dset3.removeAllElements();
	}
	if (dset4.size() > 0)
	{
	    dset4.removeAllElements();
	}
    }
    
    /**
     * add a point to the data set that is determined by the index (setNum)
     *
     * @param  p  point to be added to the input data class
     *
     */
    public void addPoint(MyPoint p)
    {
	
	
	selectFlag = 0;
	switch (set_index_d)
	{
	    
	    case 1 :    
		dset1.addElement(p);
		break;
		
	    case 2 :
		dset2.addElement(p);
		break;

	    case 3 :
		dset3.addElement(p);
		break;
	    
	    case 4 :
		dset4.addElement(p);
		break;
	}
    }
    
    /** 
     * Draws Gaussian distribution by calling setGaussian()
     *
     * @param  meanx double x value of mean point
     * @param  meany double y value of mean point
     * @param  scale DisplayScale variable used for displaying
     */
    public void drawGaussian(double meanx, double meany, DisplayScale scale )
    {
	setGaussian(points, meanx, meany, cov11, cov12, cov21, cov22, scale);
    }
    
    /**
     * creates a set of point that correspond to a gaussians distribution
     *
     * @param    points max points in the distribution
     * @param    meanx mean (x) of the distribution
     * @param    meany mean (y) of the distribution
     * @param    c11 covariance matrix element c11
     * @param    c12 covariance matrix element c12
     * @param    c21 covariance matrix element c21
     * @param    c22 covariance matrix element c22
     * @param    scale DisplayScale variable
     *
     *
     */
    public void setGaussian(
			    int points,
			    double meanx,
			    double meany,
			    double c11,
			    double c12,
			    double c21,
			    double c22,
			    DisplayScale scale)
    {

	// declare local variables
	//
	double xmax, xmin, ymax, ymin;  
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	xval = new double[points];
	yval = new double[points];

	// determine the ratio of pixels per seconds
	//
	//double xpixsec = input_width_a / (xmax - xmin);
	//double ypixsec = input_height_a / (ymax - ymin);

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(points, meanx, meany, xval, yval, c11, c12, c21, c22);

	// format the point to be plotted on the screen
	//
	//Vector dist = new Vector(points);
	Vector<MyPoint> samp = new Vector<MyPoint>(points);
	
	for (int i = 0; i < points; i++)
	{    
	    if (set_index_d == 1)
	    {
		dset1.addElement(new MyPoint(xval[i], yval[i]));
	    } 
	    else if (set_index_d == 2)
	    {
		dset2.addElement(new MyPoint(xval[i], yval[i]));
	    } 
	    else if (set_index_d == 3)
	    {
		dset3.addElement(new MyPoint(xval[i], yval[i]));
	    } 
	    else
	    {
		dset4.addElement(new MyPoint(xval[i], yval[i]));
	    }
	    
	}
	
	// for (int i = 0; i < points; i++)
	//{
	// int xpixel = (int) ((input_width_a / 2) + (xpixsec * xval[i]));
	// int ypixel = (int) ((input_width_a / 2) - (ypixsec * yval[i]));
	// dist.addElement(new Point(xpixel, ypixel));
	// samp.addElement(new MyPoint(xval[i], yval[i]));
	//}
	
	// clone the distribution to the appropriate data set
	
    }

    /**
     * create a set of point that correspond to two gaussians to
     * be displayed on the input canvas
     *
     * @param scale DisplayScale variable used for displaying
     */
    public void setTwoGaussian(DisplayScale scale)
    {

	// declare local variables
	//
	double xmax, xmin, ymax, ymin;  
	int maxsize = 0;
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	maxsize = 200    ;
	xval = new double[maxsize];
	yval = new double[maxsize];

	// generate the binormal gaussian random deviates
	//

	bn.gaussian(
		    maxsize,
		    xmin + (xmax - xmin) / 4.0,
		    ymax - (ymax - ymin) / 4.0,
		    xval,
		    yval,
		    ((0.1 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> first = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    first.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmax - (xmax - xmin) / 4,
		    ymin + (ymax - ymin) / 4,
		    xval,
		    yval,
		    ((0.1 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));
	
	// format the point to be plotted on the screen
	//
	Vector<MyPoint> second = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    second.addElement(new MyPoint(xval[i], yval[i]));
	}

	dset1 = first;
	dset2 = second;
    }

    /**
     * create a set of point that correspond to four gaussians to
     * be displayed on the input canvas
     * method: setFourGaussian
     *
     * @param scale DisplayScale variable used for displaying
     *
     */
  public void setFourGaussian(DisplayScale scale)
  {

      // local variables
      //
      double xmax, xmin, ymax, ymin;  
      int maxsize = 0;
      double[] xval = null;
      double[] yval = null;
      BiNormal bn = new BiNormal();

      // initialize variables
      //
      xmax = scale.xmax;
      xmin = scale.xmin;
      ymax = scale.ymax;
      ymin = scale.ymin;
      maxsize = 200;
      xval = new double[maxsize];
      yval = new double[maxsize];

      // generate the binormal gaussian random deviates
      //
      bn.gaussian(
		  maxsize,
		  xmax - (xmax - xmin) / 4.0,
		  ymax - (ymax - ymin) / 4.0,
		  xval,
		  yval,
		  ((0.1 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		  0.0,
		  0.0,
		  ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

      // format the point to be plotted on the screen
      //
      Vector<MyPoint> first = new Vector<MyPoint>(maxsize);
      for (int i = 0; i < maxsize; i++)
      {
	  first.addElement(new MyPoint(xval[i], yval[i]));
      }

      // generate the binormal gaussian random deviates
      //
      bn.gaussian(
		  maxsize,
		  xmin + (xmax - xmin) / 4.0,
		  ymax - (ymax - ymin) / 4.0,
		  xval,
		  yval,
		  ((0.1 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		  0.0,
		  0.0,
		  ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));
      
      // format the point to be plotted on the screen
      //
      Vector<MyPoint> second = new Vector<MyPoint>(maxsize);
      for (int i = 0; i < maxsize; i++)
      {
	  second.addElement(new MyPoint(xval[i], yval[i]));
      }
      
      // generate the binormal gaussian random deviates
      //
      bn.gaussian(
		  maxsize,
		  xmin + (xmax - xmin) / 4.0,
		  ymin + (ymax - ymin) / 4.0,
		  xval,
		  yval,
		  ((0.1 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		  0.0,
		  0.0,
		  ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

      // format the point to be plotted on the screen
      //
      Vector<MyPoint> third = new Vector<MyPoint>(maxsize);
      for (int i = 0; i < maxsize; i++)
      {
	  third.addElement(new MyPoint(xval[i], yval[i]));
      }

      // generate the binormal gaussian random deviates
      //
      bn.gaussian(
		  maxsize,
		  xmax - (xmax - xmin) / 4.0,
		  ymin + (ymax - ymin) / 4.0,
		  xval,
		  yval,
		  ((0.1 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		  0.0,
		  0.0,
		  ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

      // format the point to be plotted on the screen
      //
      Vector<MyPoint> forth = new Vector<MyPoint>(maxsize);
      for (int i = 0; i < maxsize; i++)
      {
	  forth.addElement(new MyPoint(xval[i], yval[i]));
      }
      
      dset1 = first;
      dset2 = second;
      dset3 = third;
      dset4 = forth;
  }
    
    /**
     * create a set of point that correspond to two overlapped gaussians to
     * be displayed on the input canvas
     *
     * @param scale DisplayScale variable used for displaying
     *
     */
    public void setOverGaussian(DisplayScale scale)
    {

	// local variables
	//
	double xmax, xmin, ymax, ymin;  
	int maxsize = 0;
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	maxsize = 200;
	xval = new double[maxsize];
	yval = new double[maxsize];

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    ((xmax - xmin) / 1.75 + xmin),
		    ((ymax - ymin) / 2 + ymin),
		    xval,
		    yval,
		    ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4),
		    0.0,
		    0.0,
		    ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> first = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    first.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    (xmax - (xmax - xmin) / 1.75),
		    ((ymax - ymin) / 2 + ymin),
		    xval,
		    yval,
		    ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4),
		    0.0,
		    0.0,
		    ((0.1 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> second = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    second.addElement(new MyPoint(xval[i], yval[i]));
	}
	
	dset1 = first;
	dset2 = second;
    }
    
    /**
     * create a set of point that correspond to two ellipses to
     * be displayed on the input canvas
     *
     * @param   scale DisplayScale variable used for displaying
     */
    public void setTwoEllipses(DisplayScale scale)
    {
	
	// local variables
	//
	double xmax, xmin, ymax, ymin;  
	int maxsize = 0;
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	maxsize = 200;
	xval = new double[maxsize];
	yval = new double[maxsize];

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmin + (xmax - xmin) / 4.0,
		    ymax - (ymax - ymin) / 4.0,
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> first = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    first.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmax - (xmax - xmin) / 4.0,
		    ymin + (ymax - ymin) / 4.0,
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> second = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    second.addElement(new MyPoint(xval[i], yval[i]));
	}

	dset1 = first;
	dset2 = second;
    }

    /**
     * create a set of point that correspond to two rotated ellipses to
     * be displayed on the input canvas
     *
     * @param   scale DisplayScale variable used for displaying
     */
    public void setRotatedEllipses(DisplayScale scale)  
    {

	// local variables
	//
	double xmax, xmin, ymax, ymin;  
	int maxsize = 0;
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	maxsize = 200;
	xval = new double[maxsize];
	yval = new double[maxsize];

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmin + (xmax - xmin) / 4.0,
		    ymax - (ymax - ymin) / 4.0,     
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> first = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    first.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmax - (xmax - xmin) / 4.0,
		    ymin + (ymax - ymin) / 4.0,   
		    xval,
		    yval,
		    ((0.017 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> second = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    second.addElement(new MyPoint(xval[i], yval[i]));
	}

	dset1 = first;
	dset2 = second;
    }

    /**
     * create a set of point that correspond to four ellipses to
     * be displayed on the input canvas
     *
     * @param   scale DisplayScale variable used for displaying
     *
     */
     public void setFourEllipses(DisplayScale scale)
    {

	// local variables
	//
	double xmax, xmin, ymax, ymin;  
	int maxsize = 0;
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();
	 
	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	maxsize = 200;
	xval = new double[maxsize];
	yval = new double[maxsize];
    
	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmax - (xmax - xmin) / 4.0,
		    ymax - (ymax - ymin) / 4.0,   
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> first = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    first.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmin + (xmax - xmin) / 4.0,
		    ymax - (ymax - ymin) / 4.0,
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (ymax - ymin))/4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> second = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    second.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmin + (xmax - xmin) / 4.0,
		    ymin + (ymax - ymin) / 4.0,
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> third = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    third.addElement(new MyPoint(xval[i], yval[i]));
	}

	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmax - (xmax - xmin) / 4.0,
		    ymin + (ymax - ymin) / 4.0,
		    xval,
		    yval,
		    ((0.133 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.017 * (ymax - ymin)) / 4) * (((ymax - ymin)) / 4));

	// format the point to be plotted on the screen
	//
	Vector<MyPoint> forth = new Vector<MyPoint>(maxsize);
	for (int i = 0; i < maxsize; i++)
	{
	    forth.addElement(new MyPoint(xval[i], yval[i]));
	}

	dset1 = first;
	dset2 = second;
	dset3 = third;
	dset4 = forth;
    }

    /**
     * create a set of point that correspond a toroidal to
     * be displayed on the input canvas
     *
     * @param   scale DisplayScale variable used for displaying
     *
     */
    public void setToroidal(DisplayScale scale)
    {

	// declare variables
	//
	double xmax, xmin, ymax, ymin;  
	Vector<MyPoint> torodial1;
	Vector<MyPoint> torodial2;
	int maxsize = 0;
	double[] xval = null;
	double[] yval = null;
	BiNormal bn = new BiNormal();
	double angle = 0.0;
	double radius = 0.0;
	double x1 = 0.0;
	double y1 = 0.0;
	double ring_radius;
	double stddev_radius;
	int num_ring = 2000;

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	maxsize = 400;
	xval = new double[maxsize];
	yval = new double[maxsize];
	torodial1 = new Vector<MyPoint>(50, 10);
	torodial2 = new Vector<MyPoint>(50, 10);
	ring_radius = 1.5 * ((ymax - ymin)) / 4;
	stddev_radius = 0.1 * ((ymax - ymin)) / 4;
	// make outside ring
	//
	for (int i = 1; i <= num_ring; i++)
	{
	    
	    // the angle is linearly distributed from 0 to 2 pi
	    //
	    angle = MathUtil.random.nextDouble() * 2 * Math.PI;
	    radius = MathUtil.grand(ring_radius, stddev_radius);
	    
	    // set points
	    //
	    x1 = radius * Math.cos(angle);
	    y1 = radius * Math.sin(angle);
	    
	    // plot the points
	    //	
	    torodial2.addElement(new MyPoint(x1 + xmin + (xmax-xmin)/2.0,
					     y1 + ymin + (ymax-ymin)/2.0));
	}
	
	// generate the binormal gaussian random deviates
	//
	bn.gaussian(
		    maxsize,
		    xmin + (xmax - xmin) / 2.0,
		    ymin + (ymax - ymin) / 2.0,
		    xval,
		    yval,
		    ((0.033 * (xmax - xmin)) / 4) * (((xmax - xmin)) / 4),
		    0.0,
		    0.0,
		    ((0.033 * (xmax - xmin)) / 4) * ((xmax - xmin) / 4));

	// format the point to be plotted on the screen
	//
	for (int i = 0; i < maxsize; i++)
	{
	    torodial1.addElement(new MyPoint(xval[i], yval[i]));
	}

	// initialize each torodial to each set
	//
	dset1 = torodial1;
	dset2 = torodial2;
    }

 /**
  * create a set of point that correspond a yin and yang symbol to
  * be displayed on the input canvas
  *
  * @param scale DisplayScale variable used for viewing
  *
  */
    public void setYinYang(DisplayScale scale)
    {

	// declare variables
	//
	double xmax, xmin, ymax, ymin;  
	Vector<MyPoint> Yin;
	Vector<MyPoint> Yang;
	double xpt = 0.0;
	double ypt = 0.0;
	double currPoints = 0;
	double maxPoints = 1000;
	double distance1 = 0.0;
	double distance2 = 0.0;
	double distance3 = 0.0;
	double radius1 = 0.0;
	double radius2 = 0.0;
	double stddev_center;

	// initialize variables
	//
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	stddev_center = 1.5 * (xmax - xmin) / 2;
	Yin = new Vector<MyPoint>(50, 10);
	Yang = new Vector<MyPoint>(50, 10);
	radius1 = 1.5 * (((xmax - xmin))/4);
	radius2 = 0.75 * (((xmax - xmin))/4);

	while (currPoints < maxPoints)
	{
	    
	    // generate points
	    //
	    xpt = MathUtil.grand(xmin + (xmax-xmin)/2.0, stddev_center);
	    ypt = MathUtil.grand(ymin + (ymax-ymin)/2.0, stddev_center);
	    
	    // calculate radius for each generated point
	    //
	    distance1 =
		Math.sqrt(
			  (xpt) * (xpt)
			  + (ypt) * (ypt));
	    
	    distance2 =
		Math.sqrt(
			  (xpt) * (xpt)
			  + (ypt + radius2)
			  * (ypt + radius2));
	    
	    distance3 =
		Math.sqrt(
			  (xpt) * (xpt)
			  + (ypt - radius2)
			  * (ypt - radius2));
	    
	    // decide where to plot each point
	    //
	    if (distance1 <= radius1)
	    {
		if (xpt >= (0.0 - radius1)
		    && xpt <= 0.0)
		{
		    if ((distance1 <= radius1 || distance2 <= radius2)
			&& distance3 > radius2)
		    {
			Yin.addElement(new MyPoint(xpt, ypt));
			currPoints++;
		    } 
		    else
		    {
			Yang.addElement(new MyPoint(xpt, ypt));
		    }
		}
		if (xpt > 0.0
		    && xpt <= radius1)
	       {
		   if ((distance1 <= radius1 || distance3 <= radius2)
		       && distance2 > radius2)
		   {
		       Yang.addElement(new MyPoint(xpt, ypt));
		       currPoints++;
		   } 
		   else
		  {
		      Yin.addElement(new MyPoint(xpt, ypt));
		  }
	       }
	    }
	}
	
	// set yin and yang to each set
	dset1 = Yang;
	dset2 = Yin;
    }

    /**
     * Draws a randomly generated signal.  It follows a sine curve
     * with gaussian noise
     *
     * @param scale DisplayScale variable used for viewing
     */
    public void setRandomSignal(DisplayScale scale) {

	Vector<MyPoint> set = new Vector<MyPoint>();
	Vector<MyPoint> i_set = new Vector<MyPoint>();

	// temporary process noise vector
	//
	Vector<Double> w_temp = new Vector<Double>();
	
	// temp hidden state vector
	//
	Vector<Double> x_temp = new Vector<Double>();

	// normalize process noise
	//
	w_temp = gaussianGenerator(100, 0.0, 0.4);
	
	// initial state
	//
	x_temp.addElement(0.0);
	set.addElement(new MyPoint( -0.9 , 0.0));
	
	for (int i = 2; i < 25; i++) {

	    double x = x_temp.elementAt(i - 2);
	    //	    double temp = 0.5 * x;

 	    double temp = 0.8 * (Math.sin(1 * i));
 	    temp += w_temp.elementAt(i);	    
 	    x_temp.addElement(temp);
	    temp += (25 * x) / (1 + (x * x));
	    temp += 8 * Math.cos(1.2 * (i));
	    temp += w_temp.elementAt(i);
	    
	    x_temp.addElement(temp);

	    
	    set.addElement(new MyPoint((i / 12.5 - 1.0), temp / 25.0));
	    }

	int j = 0;

	for (int i = 0; i < 23; i++) {


	    i_set.addElement(set.elementAt(i));
	    
	    // interpolate point with 3 pts (2 commented out)
	    //
	    double x = (set.elementAt(i).x + set.elementAt(i+1).x) / 2.0;
	    double y = (set.elementAt(i).y + set.elementAt(i+1).y) / 2.0;
	    MyPoint mid_pt = new MyPoint(x,y);

	    x = (set.elementAt(i).x + mid_pt.x) / 2.0;
	    y = (set.elementAt(i).y + mid_pt.y) / 2.0;
	    MyPoint first_pt = new MyPoint(x,y);

	    x = (set.elementAt(i+1).x + mid_pt.x) / 2.0;
	    y = (set.elementAt(i+1).y + mid_pt.y) / 2.0;
	    MyPoint last_pt = new MyPoint(x,y);

	    // i_set.addElement(first_pt);
	    i_set.addElement(mid_pt);
	    // i_set.addElement(last_pt);
	}
	
	dset1 = i_set;
	

    }

    /**
     * Draws a Lorentz generated signal.  It follows a sine curve
     * with gaussian noise
     *
     * @param scale DisplayScale variable used for viewing
     */
    public void setLorentzSignal(DisplayScale scale) {



// // method: computeLorentz
// //
// // arguments:
// //  VectorFloat& lorentz: (output) lorentz
// //  long channel_index: (input) channel_index
// //
// // return: a boolean value indicating status
// //
// // this method computes sample points of a chaotic series generated by a
// // Lorentz system
// //
// boolean Generator::computeLorentz(VectorFloat& lorentz_a,
// 				 long channel_index_a) {
  
//   // check if the Integrate time step is zero
//   //
//   if (chaos_integrate_upsample_d < (long)1) {
//     return Error::handle(name(), L"computeLorentz",
// 			 ERR_INTEGRATE, __FILE__, __LINE__); 
//   }

//   // check if the frequency is not zero
//   //
//   if (frequency_d == (Float)0) {
//     return Error::handle(name(), L"computeLorentz",
// 			 ERR_ZFREQ, __FILE__, __LINE__);
//   }
  
//   // Computing the length of the data-set to be generated
//   //
//   long length = (long)(Integral::round(frame_dur_d * sample_freq_d));
//   long last_frame_index
//     = (long)(Integral::ceil(signal_duration_d / frame_dur_d));

//   // set the length of the output vector
//   // Note that we will upsample the signal by 10 samples, perform the numerical
//   // integration, and then only store every 10'th sample of this upsampled signal
//   //
//   lorentz_a.setLength(length);  

//   // declare local variables
//   // sigma, r and b are coefficients of the Lorentz differential equations
//   // x1, y1, z1 are the three variables of the Lorentz differential equations
//   // which are initialized to the provided/default inital values
//   // dt is the integration time step, required for numerical integration
//   //
//   Float sigma = chaos_a_d, r = chaos_b_d, b = chaos_c_d, dt = 1.0 / ((float)chaos_integrate_upsample_d * sample_freq_d);
//   Float x1 = chaos_x0_d, y1 = chaos_y0_d, z1 = chaos_z0_d;  
//   Float dx1, dy1, dz1, dx2, dy2, dz2, dx3, dy3, dz3, dx4, dy4, dz4;
//   long i;

//   // dealing with computing last frame
//   //
//   if (accum_frame_d(channel_index_a) >= last_frame_index - 1) {
//     long last_frame_length = (long)(signal_duration_d * sample_freq_d -
// 		     frame_dur_d * sample_freq_d *
// 		     accum_frame_d(channel_index_a));

//     // Generating only as many samples as required for the last frame
//     // The following FOR loop uses Runge-Kutta Numerical Integration to generate the
//     // scalar time series from a Lorentz system
//     //
//     for (i = 0; i < last_frame_length * chaos_integrate_upsample_d; i++) {

//       dx1 = sigma * (y1 - x1) * dt;
//       dy1 = (r * x1 - y1 - x1 * z1) * dt;
//       dz1 = (-b * z1 + x1 * y1)*dt;
      
//       dx2 = sigma * ((y1 + dy1 / (float)2) - (x1 + dx1 / (float)2)) * dt;
//       dy2 = (r * (x1 + dx1 / (float)2) - (y1 + dy1 / (float)2) - (x1 + dx1 / (float)2)
// 	     * (z1 + dz1 / (float)2)) * dt;
//       dz2 = (-b * (z1 + dz1 / (float)2) + (x1 + dx1 / (float)2) *(y1 + dx1 / (float)2))*dt;
      
//       dx3 = sigma * ((y1 + dy2 / (float)3) - (x1 + dx2 / (float)3)) * dt;
//       dy3 = (r * (x1 + dx2 / (float)3) - (y1 + dy2 / (float)3) - (x1 + dx2 / (float)3)
// 	     * (z1 + dz2 / (float)3)) * dt;
//       dz3 = (-b * (z1 + dz2 / (float)3) + (x1 + dx2 / (float)3) * (y1 + dy2 / (float)3))*dt;
      
//       dx4 = sigma * ((y1 + dy3) - (x1 + dx3)) * dt;
//       dy4 = (r * (x1 + dx3) - (y1 + dy3) - (x1 + dx3) * (z1 + dz3)) * dt;
//       dz4 = (-b * (z1 + dz3) + (x1 + dx3) * (y1 + dy3)) * dt;
      
//       x1 = x1 + dx1 / (float)6 + dx2 / (float)3 + dx3 / (float)3 + dx4 / (float)6;
//       y1 = y1 + dy1 / (float)6 + dy2 / (float)3 + dy3 / (float)3 + dy4 / (float)6;
//       z1 = z1 + dz1 / (float)6 + dz2 / (float)3 + dz3 / (float)3 + dz4 / (float)6;
//       if (!(i % chaos_integrate_upsample_d)) {
// 	lorentz_a(i / chaos_integrate_upsample_d) = x1;
//       }
//     }
    
//     // set the signal points to zero for the unspecified points of
//     // last frame
//     //
//     for (long i = last_frame_length; i < length; i++) {
//       lorentz_a(i) = 0.0;
//     }    
//   }
    
//   // dealing with computing other frames
//   //
//   else {
    
//     // The following FOR loop uses Runge-Kutta Numerical Integration to generate the
//     // scalar time series from a Lorentz system
//     //
//     for (i = 0; i < length * chaos_integrate_upsample_d; i++) {
      
//       dx1 = sigma * (y1 - x1) * dt;
//       dy1 = (r * x1 - y1 - x1 * z1) * dt;
//       dz1 = (-b * z1 + x1 * y1) * dt;
      
//       dx2 = sigma * ((y1 + dy1 / (float)2) - (x1 + dx1 / (float)2)) * dt;
//       dy2 = (r * (x1 + dx1 / (float)2) - (y1 + dy1 / (float)2) - (x1 + dx1 / (float)2)
// 	     * (z1 + dz1 / (float)2)) * dt;
//       dz2 = (-b * (z1 + dz1 / (float)2) + (x1 + dx1/(float)2) * (y1 + dx1/(float)2)) * dt;
      
//       dx3 = sigma * ((y1 + dy2 / (float)3) - (x1 + dx2 / (float)3)) * dt;
//       dy3 = (r * (x1 + dx2 / (float)3) - (y1 + dy2/(float)3) - (x1 + dx2/(float)3)
// 	     * (z1 + dz2 / (float)3)) * dt;
//       dz3 = (-b * (z1 + dz2/(float)3) + (x1 + dx2/(float)3) * (y1 + dy2/(float)3)) * dt;
      
//       dx4 = sigma * ((y1 + dy3) - (x1 + dx3)) * dt;
//       dy4 = (r * (x1 + dx3) - (y1 + dy3) - (x1 + dx3) * (z1 + dz3)) * dt;
//       dz4 = (-b * (z1 + dz3) + (x1 + dx3) * (y1 + dy3)) * dt;
      
//       x1 = x1 + dx1 / (float)6 + dx2 / (float)3 + dx3 / (float)3 + dx4 / (float)6;
//       y1 = y1 + dy1 / (float)6 + dy2 / (float)3 + dy3 / (float)3 + dy4 / (float)6;
//       z1 = z1 + dz1 / (float)6 + dz2 / (float)3 + dz3 / (float)3 + dz4 / (float)6;
//       if (!(i % chaos_integrate_upsample_d)) {
// 	lorentz_a(i / chaos_integrate_upsample_d) = x1;
//       }
//     }
//     // To preserve continuity in generation, save the
//     // last samples of x,y,z as the initial conditions for the next frame.
//     //
//     chaos_x0_d = x1;
//     chaos_y0_d = y1;
//     chaos_z0_d = z1;    
//   } 

//   //  exit gracefully
//   //
//   return true;
// }






	Vector<MyPoint> set = new Vector<MyPoint>();
	Vector<MyPoint> i_set = new Vector<MyPoint>();

	// temporary process noise vector
	//
	Vector<Double> w_temp = new Vector<Double>();
	
	// temp hidden state vector
	//
	Vector<Double> x_temp = new Vector<Double>();

	// normalize process noise
	//
	w_temp = gaussianGenerator(100, 0.0, 0.4);
	
	// initial state
	//
	x_temp.addElement(0.0);
	set.addElement(new MyPoint( -0.9 , 0.0));
	
	for (int i = 2; i < 25; i++) {

	    double x = x_temp.elementAt(i - 2);
	    //	    double temp = 0.5 * x;

 	    double temp = 0.8 * (Math.sin(1 * i));
 	    temp += w_temp.elementAt(i);	    
 	    x_temp.addElement(temp);
	    temp += (25 * x) / (1 + (x * x));
	    temp += 8 * Math.cos(1.2 * (i));
	    temp += w_temp.elementAt(i);
	    
	    x_temp.addElement(temp);

	    
	    set.addElement(new MyPoint((i / 12.5 - 1.0), temp / 25.0));
	    }

	int j = 0;

	for (int i = 0; i < 23; i++) {


	    i_set.addElement(set.elementAt(i));
	    
	    // interpolate point with 3 pts (2 commented out)
	    //
	    double x = (set.elementAt(i).x + set.elementAt(i+1).x) / 2.0;
	    double y = (set.elementAt(i).y + set.elementAt(i+1).y) / 2.0;
	    MyPoint mid_pt = new MyPoint(x,y);

	    x = (set.elementAt(i).x + mid_pt.x) / 2.0;
	    y = (set.elementAt(i).y + mid_pt.y) / 2.0;
	    MyPoint first_pt = new MyPoint(x,y);

	    x = (set.elementAt(i+1).x + mid_pt.x) / 2.0;
	    y = (set.elementAt(i+1).y + mid_pt.y) / 2.0;
	    MyPoint last_pt = new MyPoint(x,y);

	    // i_set.addElement(first_pt);
	    i_set.addElement(mid_pt);
	    // i_set.addElement(last_pt);
	}
	
	dset1 = i_set;
	

    }

    /**
     * This method generates the gaussian distribution 
     *
     * @param num_samples the number of gaussian points to be generated
     * @param mean        the average value of the distribution
     * @param std_dev     the standard deviation
     *
     * @return            the vector containing the gaussian distribution
     */
    Vector<Double> gaussianGenerator(int num_samples,
				     double mean, 
				     double std_dev) {

	Vector<Double> gauss_vec = new Vector<Double> ();
	Random r = new Random();

	//std_dev /= 25.0;
	//std_dev = .00001;

	for (int i = 0; i < num_samples; i++) {

	    // nextGaussian returns a double with std_dev 1.0 and mean 0
	    // if a different std_dev or mean are desired, mult and add
	    //
	    gauss_vec.addElement((r.nextGaussian() * std_dev) + mean);
	}
 
	return gauss_vec;

    } // end of gaussian generator


    /**
     * Converts an instance of Point to an instance of MyPoint containing the
     * cartesian x-y values
     *
     * @param   jpoint Point variable used to set MyPoint
     * @param   panelWidth Width of viewing panel
     * @param   panelHeight Height of viewing panel
     * @param   scale DisplayScale variable used for viewing
     *
     * @return  MyPoint convertion
     *
     */
    static public MyPoint convertPoint(Point jpoint, int panelWidth,
				       int panelHeight, DisplayScale scale)
    {
	double xmax, xmin, ymax, ymin;
	
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	return new MyPoint(
          MathUtil.SetDecimal((((double)jpoint.x/panelWidth)
		                 * (xmax-xmin)) + xmin, 8),
	  MathUtil.SetDecimal((((panelHeight-(double)jpoint.y)/panelHeight)
				 * (ymax-ymin)) + ymin, 8));
	
    }

    /**
     * Converts a vector of Points to a vector of MyPoints containing the
     * cartesian x-y values
     *
     * @param    jset Vector of points to be converted
     * @param    panelWidth Width of viewing panel
     * @param    panelHeight Height of viewing panel
     * @return   Vector of MyPoints converted
     *
     */
    static public Vector<MyPoint> convertPoints(Vector<Point> jset, int panelWidth, int panelHeight, DisplayScale scale)
    {
	Vector<MyPoint> dset = new Vector<MyPoint>();
	
	for (int i = 1; i <= jset.size(); i++)
	{
	    dset.addElement(convertPoint((Point)jset.elementAt(i), panelWidth, panelHeight, scale));
	}
	
	return dset;
    }
    
    /**
     * Converts an instance of MyPoint to an instance of Point
     *
     * @param    dpoint MyPoint to be converted
     * @param    panelWidth Width of panel for viewing
     * @param    panelHeight Width of panel for viewing
     * @return   Point type converted from MyPoint type
     *
     */
    static public Point convertMyPoint(MyPoint dpoint, int panelWidth,
				       int panelHeight, DisplayScale scale)
    {
	double xmax, xmin, ymax, ymin;
	double xpt,ypt;
	Point jpoint;
	
	xmax = scale.xmax;
	xmin = scale.xmin;
	ymax = scale.ymax;
	ymin = scale.ymin;
	
	xpt = (panelWidth * (dpoint.x - xmin))/(xmax - xmin);
	ypt = panelHeight - (((dpoint.y-ymin) * panelHeight)/(ymax - ymin));
	jpoint = new Point((int)xpt, (int)ypt);
	
        return jpoint;
    }
    
    /**
     * Converts a Vector of MyPoints to a Vector of points that correspond to
     * Java coordinates.  
     *
     * @param   dset Vector of MyPoints to be converted to Vector of Points
     * @param   panelWidth Width of panel for viewing
     * @param   panelHeight Height of panel for viewing
     * @return  Vector of Points converted from MyPoints
     *
     */
    static public Vector<Point> convertMyPoints (Vector<MyPoint> dset, int panelWidth, int panelHeight, DisplayScale scale)
    {
	double xpt, ypt;
	Vector<Point> jset = new Vector<Point>();
	MyPoint dpoint = new MyPoint();
	
	for (int i = 0; i < dset.size(); i++)
	{
	    jset.addElement(convertMyPoint((MyPoint)dset.elementAt(i), 
					   panelWidth, panelHeight, scale));
	}
	
	return jset;
    }
	    
    /**
     * sets the gaussian values
     *
     * @param  points number of points
     * @param  cov11 double value of covariance11 
     * @param  cov12 double value of covariance12 
     * @param  cov21 double value of covariance21 
     * @param  cov22 double value of covariance22 
     *
     */    
    public void setDrawGaussValues(int points, double cov11, double cov12, 
				   double cov21, double cov22)
    {
	
	this.points = points;
	this.cov11 = cov11;
	this.cov12 = cov12;
	this.cov21 = cov21;
	this.cov22 = cov22;
    }
    
    /**
     * sets the gaussian values
     *
     * @param  color_dset1 color for data set 1
     * @param  color_dset2 color for data set 2 
     * @param  color_dset3 color for data set 3 
     * @param  color_dset4 color for data set 4
     */    
    public void setColors(Color color_dset1, Color color_dset2, 
			      Color color_dset3, Color color_dset4)
    {
	this.color_dset1 = color_dset1;
	this.color_dset2 = color_dset2;
	this.color_dset3 = color_dset3;
	this.color_dset4 = color_dset4;
    }  

}
