/**
 * file: InputPanel.java
 *
 * last edited: Ryan Irwin
 */

// import necessary java libraries
//
import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.*;

//These imports are not needed - Phil T. 6-23-03
//import java.awt.event.*;

/**
 * class acts as a top level container that emcompasses the input plot
 *
 * hierarchy: JPanel->SubPanel->InputPanel
 *
 */
public class InputPanel extends SubPanel 
{
    
    // *********************************************************************
    //
    // declare global variables and components
    //
    // *********************************************************************

    // declare components
    //
    DataPoints data_points_d;

    String title;

    // selection area for input display
    //
    SelectionArea selection_area_d;  

    // make display of input panel
    //          
    DisplayArea disp_area_d;

    Font curr_font_d = getFont();   // fonts for the buttons

    Font new_font_d = new Font(curr_font_d.getName(), 
			    curr_font_d.getStyle(), 12);
  
    Font currentFont = getFont(); // fonts for the buttons
    Font newFont = new Font(currentFont.getName(), currentFont.getStyle(), 12);
    int width = 0;                  // width of the canvas
    int height = 0;                 // height of the canvas
    
    int zeroX = 0;                  // zero position on x axis
    int zeroY = 0;                  // zero position on y axis

    // *********************************************************************
    //
    // declare class constructors
    //
    // *********************************************************************
   
    /**
     * constructor sets the border and initializes the input plot
     *
     * @param   data_a input data samples
     *
     */
    InputPanel(DataPoints data_a) 
    {
	
	super();

	// initialize the panel used for plotting
	//
	disp_area_d = new DisplayArea();

	// initialize components
	//
	data_points_d = data_a;

	// register a listener for the radio buttons.
	//
	selection_area_d = new SelectionArea(data_points_d);

	// set the border for the panel
	//
	String title = new String("Input Display:");

	Border titleBorder = BorderFactory.createTitledBorder(
			     BorderFactory.createLineBorder(Color.black), 
			     title, 
			     TitledBorder.LEFT, 
			     TitledBorder.ABOVE_TOP, 
			     new_font_d, 
			     Color.black);
	setBorder(titleBorder);
    }


    /**
     * constructor initializes the samples to be plotted in the signal panel
     *
     */
    InputPanel()
    {

	super();

	// initialize the panel used for plotting
	//
	disp_area_d = new DisplayArea();

	// set the border for the panel
	//
	title = new String("Input Display:");
	Border titleBorder =
	    BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.black),
				title,
				TitledBorder.LEFT,
				TitledBorder.ABOVE_TOP,
				newFont,
				Color.black);
	setBorder(titleBorder);
    }

    /**
     * constructor initializes the samples to be plotted in the signal panel
     *
     * @param   toCopy InputPanel to be initialized
     */
    InputPanel(InputPanel toCopy)
    {

	super();

	// initialize the panel used for plotting
	//
	disp_area_d = toCopy.disp_area_d;

	// set the border for the panel
	//
	title = new String("Input Display:");
	Border titleBorder =
	    BorderFactory.createTitledBorder(
				BorderFactory.createLineBorder(Color.black),
				title,
				TitledBorder.LEFT,
				TitledBorder.ABOVE_TOP,
				newFont,
				Color.black);
	setBorder(titleBorder);
    }


    // *********************************************************************
    //
    // declare class methods
    //
    // *********************************************************************
   
    /**
     * set the value of the draw flag for the input canvas
     *
     * @param  flag input draw flag
     *
     */
    public void setDrawFlag(boolean flag) 
    {
	selection_area_d.drawFlag = flag;
    }

    /**
     * add components to the input panel
     *
     */
    public void add_components() 
    {

	constrain(this, selection_area_d, 0, 0, GridBagConstraints.REMAINDER, 
		  GridBagConstraints.REMAINDER, GridBagConstraints.BOTH,
		  GridBagConstraints.NORTHWEST, 1, 1, 2, 2, 2, 2);
    }
    
    /**
     * redraw the input panel screen. Calls repaint()
     *
     */
    public void refresh() 
    {
	super.repaint();
	selection_area_d.repaint();
    }
    
    /**    
     * redraw the input panel screen. Calls repaint()
     *
     */
    public void clear() 
    {
	super.repaint();
	selection_area_d.clear();
	selection_area_d.repaint();
    }
  
    /**
     * Adds points to the input panel
     *
     * @param points_a points to be added to panel
     * @param type_a integer value of type
     * @param color_a color of point
     *
     * @return   true
     *
     */
    public boolean addInput(Vector<MyPoint> points_a, int type_a, Color color_a)    {

	disp_area_d.input_points_d.addElement(points_a);
	disp_area_d.input_type_d.addElement(new Integer(type_a));
	disp_area_d.input_color_d.addElement(color_a);
	
	// init the algorithm
	//
	return true;
    }   

    /*
    /**
     * set the canvas dimensions, i.e., the input plot dimensions
     *
     */
    /*
      public void setCanvasDimensions() {
      data.inputWidth = slection_area_d.getCanvasWidth();
      data.inputHeight = slection_area_d.getCanvasHeight();
      }
    */
    
    /**
     * determine the dimensions of the selection area
     *
     */
    public void DetermineDimensions() 
    {

	Dimension dimen = this.getSize();

	width = dimen.width;
	height = dimen.height;

	// set the coordinates of the axis origin
	//
	zeroX = width >> 1;
	zeroY = height >> 1;

	// store the dimensions of the output canvas
	//
	//setCanvasDimensions();
    }

    /**
     * paing the current data points and if needed the decision regions
     *
     * @param   g graphics object to be 
     */
     public void paintComponent(Graphics g) 
    {
	super.paintComponent(g);
    }

    /**
     * Duplicates the InputPanel Object
     *
     * @return  clone of InputPanel
     */
    public InputPanel copy()
    {
	try 
	{
	    return (InputPanel)clone();
	}
	catch (CloneNotSupportedException e) 
	    {}

	return null;	  
    }
}
